"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var runtime_1 = require("./runtime");
//------------------------------------------------------------------------
// Utils
//------------------------------------------------------------------------
function isResolved(field) {
    return !!field || field === 0;
}
function sumTimes(roundArray, transaction, round) {
    if (!roundArray)
        return 0;
    var total = 0;
    for (var _i = 0, roundArray_1 = roundArray; _i < roundArray_1.length; _i++) {
        var cur = roundArray_1[_i];
        if (Math.abs(cur) - 1 <= round) {
            total += cur > 0 ? 1 : -1;
        }
    }
    return total;
}
var HashIndex = (function () {
    function HashIndex() {
        this.eavIndex = runtime_1.createHash();
        this.aveIndex = runtime_1.createHash();
        this.cardinality = 0;
    }
    HashIndex.prototype.getOrCreateHash = function (parent, key) {
        var found = parent[key];
        if (!found) {
            found = parent[key] = runtime_1.createHash("hashLevel");
        }
        return found;
    };
    HashIndex.prototype.getOrCreateArray = function (parent, key) {
        var found = parent[key];
        if (!found) {
            found = parent[key] = runtime_1.createArray("hashVix");
        }
        return found;
    };
    HashIndex.prototype.roundArrayInsert = function (arr, change) {
        var round = change.round + 1;
        var neue = round * change.count;
        var ix = 0;
        var handled = false;
        for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
            var cur = arr_1[_i];
            var curRound = Math.abs(cur);
            if (curRound === round) {
                var updated = cur + neue;
                if (updated === 0) {
                    arr.splice(ix, 1);
                }
                else {
                    arr[ix] = updated;
                }
                handled = true;
                break;
            }
            else if (curRound > round) {
                arr.splice(ix, 0, neue);
                handled = true;
                break;
            }
            ix++;
        }
        if (!handled)
            arr.push(neue);
    };
    HashIndex.prototype.insert = function (change) {
        var _a = this, getOrCreateHash = _a.getOrCreateHash, getOrCreateArray = _a.getOrCreateArray;
        var eIx = getOrCreateHash(this.eavIndex, change.e);
        var aIx = getOrCreateHash(eIx, change.a);
        var vIx = getOrCreateArray(aIx, change.v);
        this.roundArrayInsert(vIx, change);
        var shouldRemove = false;
        if (!vIx.length) {
            this.cardinality--;
            delete aIx[change.v];
            if (!Object.keys(aIx).length) {
                delete eIx[change.a];
                if (!Object.keys(eIx).length) {
                    delete this.eavIndex[change.e];
                }
            }
            shouldRemove = true;
        }
        aIx = getOrCreateHash(this.aveIndex, change.a);
        vIx = getOrCreateHash(aIx, change.v);
        eIx = getOrCreateArray(vIx, change.e);
        if (shouldRemove) {
            delete vIx[change.e];
            if (!Object.keys(vIx).length) {
                delete aIx[change.v];
                if (!Object.keys(aIx).length) {
                    delete this.aveIndex[change.a];
                }
            }
        }
        else {
            this.roundArrayInsert(eIx, change);
            this.cardinality++;
        }
    };
    HashIndex.prototype.resolveProposal = function (proposal) {
        return proposal.info;
    };
    HashIndex.prototype.propose = function (proposal, e, a, v, n, transaction, round) {
        var forFields = proposal.forFields;
        forFields.clear();
        if (isResolved(e)) {
            return this.walkPropose(proposal, this.eavIndex, e, a, v, n, "a", "v", transaction, round);
        }
        else if (isResolved(a)) {
            return this.walkPropose(proposal, this.aveIndex, a, v, e, n, "v", "e", transaction, round);
        }
        else {
            // propose for attribute since that's likely to be the smallest
            forFields.push("a");
            proposal.info = Object.keys(this.aveIndex);
            proposal.cardinality = proposal.info.length;
        }
        return proposal;
    };
    HashIndex.prototype.walkPropose = function (proposal, index, a, b, c, n, fieldB, fieldC, transaction, round) {
        var forFields = proposal.forFields;
        forFields.clear();
        var bIx = index[a];
        if (!bIx) {
            proposal.cardinality = 0;
            return proposal;
        }
        if (isResolved(b)) {
            var cIx = bIx[b];
            if (!cIx) {
                proposal.cardinality = 0;
                return proposal;
            }
            if (isResolved(c)) {
                var roundArray = cIx[c];
                if (roundArray) {
                    proposal.skip = true;
                    return proposal;
                }
                proposal.cardinality = 0;
                return proposal;
            }
            else {
                forFields.push(fieldC);
                proposal.info = Object.keys(cIx);
                proposal.cardinality = proposal.info.length;
                return proposal;
            }
        }
        else {
            forFields.push(fieldB);
            proposal.info = Object.keys(bIx);
            proposal.cardinality = proposal.info.length;
            return proposal;
        }
    };
    // This function checks that there is at least one value in the index that matches the
    // given pattern. If a level is free, we have to run through the potential values
    // until we come across one that could match or we run out of values to check.
    HashIndex.prototype.walkCheck = function (index, a, b, c, n, transaction, round) {
        var bIx = index[a];
        if (!bIx)
            return false;
        if (isResolved(b)) {
            var cIx = bIx[b];
            if (!cIx)
                return false;
            if (isResolved(c)) {
                var roundArray = cIx[c];
                if (roundArray) {
                    return true;
                }
                return false;
            }
            else {
                return Object.keys(cIx).length !== 0;
            }
        }
        else {
            for (var _i = 0, _a = Object.keys(bIx); _i < _a.length; _i++) {
                var key = _a[_i];
                var cIx = bIx[key];
                if (!cIx)
                    continue;
                if (isResolved(c)) {
                    var roundArray = cIx[c];
                    if (roundArray) {
                        return true;
                    }
                    return false;
                }
                else if (Object.keys(cIx).length !== 0) {
                    return true;
                }
            }
        }
        return false;
    };
    HashIndex.prototype.check = function (e, a, v, n, transaction, round) {
        if (isResolved(e)) {
            return this.walkCheck(this.eavIndex, e, a, v, n, transaction, round);
        }
        else if (isResolved(a)) {
            return this.walkCheck(this.aveIndex, a, v, e, n, transaction, round);
        }
        return true;
    };
    // This function finds all EAVs in the index that match the given
    // pattern at the stated time. If a level is free, we have to run
    // through the potential values until we come across one that could
    // match or we run out of values to check.
    HashIndex.prototype.walkGet = function (index, a, b, c, n, fieldB, fieldC, transaction, round) {
        var fieldA = "e";
        if (fieldB === "e")
            fieldA = "a";
        var results = runtime_1.createArray("IndexWalkGet");
        var bIx = index[a];
        if (!bIx)
            return results;
        if (isResolved(b)) {
            var cIx = bIx[b];
            if (!cIx)
                return results;
            if (isResolved(c)) {
                if (sumTimes(cIx[c], transaction, round) > 0) {
                    results.push((_a = {}, _a[fieldA] = +a, _a[fieldB] = +b, _a[fieldC] = +c, _a.n = n, _a));
                }
                return results;
            }
            else {
                for (var _i = 0, _b = Object.keys(cIx); _i < _b.length; _i++) {
                    var c_1 = _b[_i];
                    if (sumTimes(cIx[c_1], transaction, round) > 0) {
                        results.push((_c = {}, _c[fieldA] = +a, _c[fieldB] = +b, _c[fieldC] = +c_1, _c.n = n, _c));
                    }
                }
                return results;
            }
        }
        else {
            for (var _d = 0, _e = Object.keys(bIx); _d < _e.length; _d++) {
                var b_1 = _e[_d];
                var cIx = bIx[b_1];
                if (!cIx)
                    continue;
                if (isResolved(c)) {
                    if (sumTimes(cIx[c], transaction, round) > 0) {
                        results.push((_f = {}, _f[fieldA] = +a, _f[fieldB] = +b_1, _f[fieldC] = +c, _f.n = n, _f));
                    }
                }
                else {
                    for (var _g = 0, _h = Object.keys(cIx); _g < _h.length; _g++) {
                        var c_2 = _h[_g];
                        if (sumTimes(cIx[c_2], transaction, round) > 0) {
                            results.push((_j = {}, _j[fieldA] = +a, _j[fieldB] = +b_1, _j[fieldC] = +c_2, _j.n = n, _j));
                        }
                    }
                }
            }
            return results;
        }
        var _a, _c, _f, _j;
        // throw new Error("HashIndex.walkGet eav not implemented.");
    };
    HashIndex.prototype.get = function (e, a, v, n, transaction, round) {
        if (isResolved(e)) {
            return this.walkGet(this.eavIndex, e, a, v, n, "a", "v", transaction, round);
        }
        else if (isResolved(a)) {
            return this.walkGet(this.aveIndex, a, v, e, n, "v", "e", transaction, round);
        }
        else
            throw new Error("HashIndex.get eaV not implemented.");
    };
    HashIndex.prototype.getDiffs = function (e, a, v, n) {
        var aIx = this.eavIndex[e];
        if (aIx) {
            var vIx = aIx[a];
            if (vIx && vIx[v]) {
                return vIx[v];
            }
        }
        return runtime_1.createArray();
    };
    return HashIndex;
}());
exports.HashIndex = HashIndex;
//------------------------------------------------------------------------
// Bit Matrix index
//------------------------------------------------------------------------
var BitMatrixIndex = (function () {
    function BitMatrixIndex() {
    }
    BitMatrixIndex.prototype.insert = function (change) {
        throw new Error("not implemented");
    };
    BitMatrixIndex.prototype.propose = function (proposal, e, a, v, n, transaction, round) {
        throw new Error("not implemented");
    };
    BitMatrixIndex.prototype.resolveProposal = function (proposal) {
        throw new Error("not implemented");
    };
    BitMatrixIndex.prototype.check = function (e, a, v, n, transaction, round) {
        throw new Error("not implemented");
    };
    BitMatrixIndex.prototype.getDiffs = function (e, a, v, n) {
        throw new Error("not implemented");
    };
    BitMatrixIndex.prototype.get = function (e, a, v, n, transaction, round) {
        throw new Error("not implemented");
    };
    return BitMatrixIndex;
}());
exports.BitMatrixIndex = BitMatrixIndex;
//------------------------------------------------------------------------
// DistinctIndex
//------------------------------------------------------------------------
var DistinctIndex = (function () {
    function DistinctIndex() {
        this.index = {};
    }
    DistinctIndex.prototype.getDelta = function (last, next) {
        var delta = 0;
        if (last == 0 && next > 0)
            delta = 1;
        if (last > 0 && next == 0)
            delta = -1;
        if (last > 0 && next < 0)
            delta = -1;
        if (last < 0 && next > 0)
            delta = 1;
        return delta;
    };
    DistinctIndex.prototype.shouldOutput = function (key, prefixRound, prefixCount) {
        var index = this.index;
        var roundCounts = index[key] || runtime_1.createArray("Insert intermediate counts");
        index[key] = roundCounts;
        var curCount = 0;
        var startingCount = roundCounts[prefixRound] = roundCounts[prefixRound] || 0;
        var minRound = Math.min(roundCounts.length, prefixRound + 1);
        for (var roundIx = 0; roundIx < minRound; roundIx++) {
            var prevCount = roundCounts[roundIx];
            if (!prevCount)
                continue;
            curCount += prevCount;
        }
        var deltas = [];
        // We only need delta changed here because if there's a round delta, it
        // would have been applied already.
        if (prefixCount === -Infinity) {
            if (curCount === Infinity) {
                curCount = 1;
                startingCount = 1;
            }
            prefixCount = -curCount;
        }
        var nextCount = curCount + prefixCount;
        var delta = this.getDelta(curCount, nextCount);
        if (delta) {
            deltas.push(prefixRound, delta);
        }
        curCount = nextCount;
        roundCounts[prefixRound] = startingCount + prefixCount;
        for (var roundIx = prefixRound + 1; roundIx < roundCounts.length; roundIx++) {
            var roundCount = roundCounts[roundIx];
            if (roundCount === undefined || roundCount === 0)
                continue;
            var lastCount = curCount - prefixCount;
            var nextCount_1 = lastCount + roundCount;
            var delta_1 = this.getDelta(lastCount, nextCount_1);
            var lastCountChanged = curCount;
            var nextCountChanged = curCount + roundCount;
            var deltaChanged = this.getDelta(lastCountChanged, nextCountChanged);
            // let finalDelta = deltaChanged - delta;
            var finalDelta = 0;
            if (delta_1 && delta_1 !== deltaChanged) {
                // undo delta
                finalDelta = -delta_1;
            }
            else if (delta_1 !== deltaChanged) {
                finalDelta = deltaChanged;
            }
            if (finalDelta) {
                deltas.push(roundIx, finalDelta);
            }
            curCount = nextCountChanged;
        }
        return deltas;
    };
    DistinctIndex.prototype.distinct = function (input, results) {
        var e = input.e, a = input.a, v = input.v, n = input.n, round = input.round, count = input.count;
        // @FIXME: When we start to unintern, we'll have invalid keys left in this index,
        // so we'll need to delete the keys themselves from the index
        var key = e + "|" + a + "|" + v;
        var deltas = this.shouldOutput(key, round, count);
        for (var deltaIx = 0; deltaIx < deltas.length; deltaIx += 2) {
            var deltaRound = deltas[deltaIx];
            var delta = deltas[deltaIx + 1];
            var change = new runtime_1.Change(e, a, v, n, input.transaction, deltaRound, delta);
            results.push(change);
        }
        return deltas.length > 0;
    };
    DistinctIndex.prototype.distinctKey = function (key, round, count, results) {
        var deltas = this.shouldOutput(key, round, count);
        for (var deltaIx = 0; deltaIx < deltas.length; deltaIx += 2) {
            var deltaRound = deltas[deltaIx];
            var delta = deltas[deltaIx + 1];
            results.push([deltaRound, delta]);
        }
        return deltas.length > 0;
    };
    DistinctIndex.prototype.getCounts = function (change) {
        var e = change.e, a = change.a, v = change.v;
        var key = e + "|" + a + "|" + v;
        return this.index[key];
    };
    DistinctIndex.prototype.sanityCheck = function () {
        var failed = false;
        var index = this.index;
        for (var key in index) {
            var counts = index[key];
            var sum = 0;
            for (var _i = 0, counts_1 = counts; _i < counts_1.length; _i++) {
                var c = counts_1[_i];
                if (!c)
                    continue;
                sum += c;
                if (sum < 0) {
                    failed = true;
                    console.error("# Negative postDistinct: ", key, counts.slice());
                }
            }
        }
        if (failed)
            throw new Error("Distinct sanity check failed.");
    };
    return DistinctIndex;
}());
exports.DistinctIndex = DistinctIndex;
//# sourceMappingURL=indexes.js.map