"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
// import {Program} from "./dsl2";
var globalsToTrack = ["transaction"];
var propertiesToTrack = ["block", "PresolveCheck", "GenericJoin"];
var PerformanceTracker = (function () {
    function PerformanceTracker() {
        this.reset();
        this.now = exports.now;
        this.elapsed = exports.elapsed;
    }
    PerformanceTracker.prototype._makePropertyHolder = function (props) {
        if (props === void 0) { props = propertiesToTrack; }
        var neue = {};
        for (var _i = 0, props_1 = props; _i < props_1.length; _i++) {
            var property = props_1[_i];
            neue[property] = 0;
        }
        return neue;
    };
    PerformanceTracker.prototype.reset = function () {
        this.activeBlock = "";
        this.activeProperties = {};
        this.times = this._makePropertyHolder(globalsToTrack);
        this.counts = this._makePropertyHolder(globalsToTrack);
        this.blocks = {};
    };
    PerformanceTracker.prototype.block = function (name) {
        var blocks = this.blocks;
        var found = blocks[name];
        if (!found) {
            found = blocks[name] = { counts: this._makePropertyHolder(), times: this._makePropertyHolder() };
        }
        this.activeBlock = name;
        this.activeProperties["block"] = this.now();
        found.counts["block"]++;
    };
    PerformanceTracker.prototype.blockEnd = function (name) {
        var _a = this, blocks = _a.blocks, activeBlock = _a.activeBlock;
        blocks[activeBlock].times["block"] += this.elapsed(this.activeProperties["block"]);
        this.activeBlock = "";
    };
    PerformanceTracker.prototype.blockTime = function (property) {
        var _a = this, blocks = _a.blocks, activeBlock = _a.activeBlock;
        var found = blocks[activeBlock];
        this.activeProperties[property] = this.now();
        found.counts[property]++;
    };
    PerformanceTracker.prototype.blockTimeEnd = function (property) {
        var _a = this, blocks = _a.blocks, activeBlock = _a.activeBlock;
        var found = blocks[activeBlock];
        found.times[property] += this.elapsed(this.activeProperties[property]);
    };
    PerformanceTracker.prototype.time = function (property) {
        var counts = this.counts;
        this.activeProperties[property] = this.now();
        counts[property]++;
    };
    PerformanceTracker.prototype.timeEnd = function (property) {
        var times = this.times;
        times[property] += this.elapsed(this.activeProperties[property]);
    };
    PerformanceTracker.prototype.serialize = function () {
        return JSON.stringify({
            times: this.times,
            counts: this.counts,
            blocks: this.blocks
        });
    };
    return PerformanceTracker;
}());
exports.PerformanceTracker = PerformanceTracker;
var NoopPerformanceTracker = (function (_super) {
    __extends(NoopPerformanceTracker, _super);
    function NoopPerformanceTracker() {
        var _this = _super.call(this) || this;
        _this.now = function () { return 0; };
        _this.elapsed = function (start) { return 0; };
        return _this;
    }
    NoopPerformanceTracker.prototype.reset = function () { };
    NoopPerformanceTracker.prototype.time = function (property) { };
    NoopPerformanceTracker.prototype.timeEnd = function (property) { };
    NoopPerformanceTracker.prototype.block = function (name) { this.activeBlock = name; };
    NoopPerformanceTracker.prototype.blockEnd = function (name) { this.activeBlock = ""; };
    NoopPerformanceTracker.prototype.blockTime = function (property) { };
    NoopPerformanceTracker.prototype.blockTimeEnd = function (property) { };
    return NoopPerformanceTracker;
}(PerformanceTracker));
exports.NoopPerformanceTracker = NoopPerformanceTracker;
if (global.process) {
    exports.now = function (start) {
        return process.hrtime();
    };
    exports.elapsed = function (start) {
        var end = process.hrtime(start);
        return ((end[0] * 1000) + (end[1] / 1000000));
    };
}
else {
    exports.now = function (start) {
        return performance.now();
    };
    exports.elapsed = function (start) {
        var end = performance.now();
        return end - start;
    };
}
//# sourceMappingURL=performance.js.map