"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var indexes_1 = require("./indexes");
var trace_1 = require("./trace");
//------------------------------------------------------------------------
// debugging utilities
//------------------------------------------------------------------------
var TRACE = false;
// Turning this on causes all of the debug(.*) statements to print to the
// console.  This is useful to see exactly what the runtime is doing as it
// evaluates a transaction, but it incurs both a pretty serious performance
// cost and prints a lot of stuff.
var DEBUG = false;
exports.debug = function () { };
if (DEBUG) {
    exports.debug = function () {
        console.log.apply(console, arguments);
    };
}
function indent(text, level) {
    var padding = new Array(level + 1).join(" ");
    return text.split("\n").join("\n" + padding);
}
function printField(field) {
    if (isRegister(field))
        return "[" + field.offset + "]";
    if (field === undefined || field === null)
        return field;
    var raw = maybeReverse(field);
    return typeof raw === "string" ? "\"" + raw + "\"" : raw;
}
exports.printField = printField;
function printFieldArray(fields) {
    return "[" + fields.map(printField).join(", ") + "]";
}
exports.printFieldArray = printFieldArray;
function printPrefix(prefix) {
    return prefix.map(function (v) { return exports.GlobalInterner.reverse(v); });
}
exports.printPrefix = printPrefix;
function toString(x) {
    if (x && x.toString)
        return x.toString();
    console.warn("No toString specified for", x);
    return "";
}
function printBlock(block) {
    return block.toString();
}
exports.printBlock = printBlock;
global.printBlock = printBlock;
function maybeReverse(value) {
    if (value === undefined)
        return value;
    var raw = exports.GlobalInterner.reverse(value);
    return ("" + raw).indexOf("|") === -1 ? raw : value;
}
exports.maybeReverse = maybeReverse;
//------------------------------------------------------------------------
// Allocations
//------------------------------------------------------------------------
// As this is a language runtime, we want to get insight into how we're using
// memory and what allocation costs we're eating as we run. To track that, we
// use createHash and createArray to give us some rough numbers. The JIT will
// inline these functions, so the cost over just using {} or [], is fairly
// negligible. In a release build we can also strip the allocation tracking.
exports.ALLOCATION_COUNT = {};
function createHash(place) {
    if (place === void 0) { place = "unknown-hash"; }
    if (!exports.ALLOCATION_COUNT[place])
        exports.ALLOCATION_COUNT[place] = 0;
    exports.ALLOCATION_COUNT[place]++;
    return Object.create(null);
}
exports.createHash = createHash;
function createArray(place) {
    if (place === void 0) { place = "unknown"; }
    if (!exports.ALLOCATION_COUNT[place])
        exports.ALLOCATION_COUNT[place] = 0;
    exports.ALLOCATION_COUNT[place]++;
    return [];
}
exports.createArray = createArray;
function copyArray(arr, place) {
    if (place === void 0) { place = "unknown"; }
    if (!exports.ALLOCATION_COUNT[place])
        exports.ALLOCATION_COUNT[place] = 0;
    exports.ALLOCATION_COUNT[place]++;
    return arr.slice();
}
exports.copyArray = copyArray;
function copyHash(hash, place) {
    if (place === void 0) { place = "unknown"; }
    if (!exports.ALLOCATION_COUNT[place])
        exports.ALLOCATION_COUNT[place] = 0;
    exports.ALLOCATION_COUNT[place]++;
    var neue = {};
    for (var _i = 0, _a = Object.keys(hash); _i < _a.length; _i++) {
        var key = _a[_i];
        neue[key] = hash[key];
    }
    return neue;
}
exports.copyHash = copyHash;
// given two arrays, append the second's items on to the first
function concatArray(arr, arr2) {
    var ix = arr.length;
    for (var _i = 0, arr2_1 = arr2; _i < arr2_1.length; _i++) {
        var elem = arr2_1[_i];
        arr[ix] = elem;
        ix++;
    }
    return arr;
}
exports.concatArray = concatArray;
// overwrite the first array with the values of the second array
// and fix the length if it's different
function moveArray(arr, arr2) {
    var ix = 0;
    for (var _i = 0, arr_1 = arr; _i < arr_1.length; _i++) {
        var elem = arr_1[_i];
        arr2[ix] = arr[ix];
    }
    if (arr2.length !== arr.length)
        arr2.length = arr.length;
    return arr2;
}
exports.moveArray = moveArray;
//------------------------------------------------------------------------
// Iterator
//------------------------------------------------------------------------
// To reduce allocations as much as possible, we want to reuse arrays as much
// as possible. If we reused the array by setting its length to 0 or to some
// new size that is smaller than its current length, we eat the cost of
// deallocating some chunk of memory as well as the potential cost in
// fragmentation. Instead, the Iterator class never changes the size of its
// backing array, and instead keeps its own length. You iterate through the
// array using the next() method:
//
// let current;
// while((current = iterator.next()) !== undefined) {
//   ...
// }
//
// Through the magic of the JIT, this has no performance penalty over using a
// standard for loop. You can get some of those "zero-cost abstractions" in JS
// too!
var Iterator = (function () {
    function Iterator() {
        this.array = [];
        this.length = 0;
        this.ix = 0;
    }
    Iterator.prototype.push = function (value) {
        this.array[this.length++] = value;
    };
    Iterator.prototype.clear = function () {
        this.length = 0;
        this.reset();
    };
    Iterator.prototype.reset = function () {
        this.ix = 0;
    };
    Iterator.prototype.next = function () {
        if (this.ix < this.length)
            return this.array[this.ix++];
        return;
    };
    Iterator.prototype.iter = function () {
        return new ReadOnlyIterator(this.array, this.length);
    };
    return Iterator;
}());
exports.Iterator = Iterator;
var ReadOnlyIterator = (function (_super) {
    __extends(ReadOnlyIterator, _super);
    function ReadOnlyIterator(arr, length) {
        var _this = _super.call(this) || this;
        _this.array = arr;
        _this.length = length;
        return _this;
    }
    ReadOnlyIterator.prototype.push = function (value) {
        throw new Error("Cannot write to a readonly iterator");
    };
    return ReadOnlyIterator;
}(Iterator));
exports.ReadOnlyIterator = ReadOnlyIterator;
function isNumber(thing) {
    return typeof thing === "number";
}
var Interner = (function () {
    function Interner() {
        // IDs are only positive integers so that they can be used as array indexes
        // for efficient lookup.
        this.currentID = 0;
        // We currently only have two value types in Eve at the moment, strings and
        // numbers.  Because keys in a javascript object are always converted to
        // strings, we have to keep dictionaries for the two types separate,
        // otherwise the number 1 and the string "1" would end up being the same
        // value;
        this.strings = createHash();
        this.numbers = createHash();
        // We use this array as a lookup from an integer ID to a RawValue since the
        // IDs are guaranteed to be densely packed, this gives us much better
        // performance than using another hash.
        this.IDs = createArray();
        // This is used as another lookup from ID to the number of references this ID
        // has in the system. As the ref count goes to zero, we can add the ID to the
        // free list so that it can be reused.
        this.IDRefCount = createArray();
        this.IDFreeList = createArray();
        // During the course of evaluation, we might allocate a bunch of intermediate
        // IDs whose values might just be thrown away. For example if we generate a
        // value just to use as a filter, there's no sense in us keeping the value in
        // the interned space.  Arenas are named groups of allocations that we may
        // want to dereference all together.  Note that just because we may
        // dereference it once, that doesn't mean the ID should be released - other
        // uses of the ID may exist.
        this.arenas = createHash();
        // The only arena we *know* we want from the beginning is for the output of functions.
        this.arenas["functionOutput"] = new Iterator();
    }
    Interner.prototype._getFreeID = function () {
        return this.IDFreeList.pop() || this.currentID++;
    };
    Interner.prototype.reference = function (id) {
        this.IDRefCount[id]++;
    };
    // Intern takes a value and gives you the ID associated with it. If there isn't an
    // ID it should create one for this value and in either case it should add a reference.
    Interner.prototype.intern = function (value) {
        var coll;
        if (isNumber(value)) {
            coll = this.numbers;
        }
        else {
            coll = this.strings;
        }
        var found = coll[value];
        if (found === undefined) {
            found = this._getFreeID();
            coll[value] = found;
            this.IDs[found] = value;
            this.IDRefCount[found] = 1;
        }
        else {
            this.IDRefCount[found]++;
        }
        return found;
    };
    // Get neither creates an ID nor adds a reference to the ID, it only looks up the
    // ID for a value if it exists.
    Interner.prototype.get = function (value) {
        var coll;
        if (isNumber(value)) {
            coll = this.numbers;
        }
        else {
            coll = this.strings;
        }
        return coll[value];
    };
    // Go from an ID to the RawValue
    Interner.prototype.reverse = function (id) {
        return this.IDs[id];
    };
    // Dereference an ID and if there are no remaining references, add it to the freelist.
    Interner.prototype.release = function (id) {
        if (id === undefined)
            return;
        this.IDRefCount[id]--;
        if (!this.IDRefCount[id]) {
            var value = this.IDs[id];
            var coll = void 0;
            if (isNumber(value)) {
                coll = this.numbers;
            }
            else {
                coll = this.strings;
            }
            coll[value] = undefined;
            this.IDs[id] = undefined;
            this.IDFreeList.push(id);
        }
    };
    Interner.prototype.arenaIntern = function (arenaName, value) {
        // @FIXME: Unfortunately we can't use arena intern at the moment due to the
        // fact that while we can know what values end up in the primary indexes,
        // we don't know what values might be hiding in intermediate indexes that
        // runtime nodes sometimes need to keep. If we *did* deallocate an arena
        // and the value didn't make it to a primary index, but ended up in an
        // intermediate one, we'd have effectively corrupted our program. The ID
        // would be freed, and then used for some completely different value. Until
        // we can find an accurate (and cheap!) way to track what values are still
        // hanging around, we'll just have to eat the cost of interning all the
        // values we've seen. Keep in mind that this isn't as bad as it sounds, as
        // the only values that would actually be freed this way are values that
        // are calculated but never end up touching the primary indexes. This is
        // rare enough that in practice, this probably isn't a big deal.
        throw new Error("Arena interning isn't ready for primetime yet.");
        // let arena = this.arenas[arenaName];
        // if(!arena) {
        //   arena = this.arenas[arenaName] = new Iterator<ID>();
        // }
        // // @NOTE: for performance reasons it might make more sense to prevent duplicates
        // // from ending up in the list. If that's the case, we could either keep a seen
        // // hash or do a get and only intern if it hasn't been seen. This is (probably?)
        // // a pretty big performance gain in the case where a bunch of rows might repeat
        // // the same function output over and over.
        // let id = this.intern(value);
        // arena.push(id);
        // return id;
    };
    Interner.prototype.releaseArena = function (arenaName) {
        var arena = this.arenas[arenaName];
        if (!arena) {
            console.warn("Trying to release unknown arena: " + arenaName);
            return;
        }
        var id;
        while ((id = arena.next()) !== undefined) {
            this.release(id);
        }
        arena.clear();
    };
    return Interner;
}());
exports.Interner = Interner;
// The runtime uses a single global interner so that all values remain comparable.
exports.GlobalInterner = new Interner();
global["GlobalInterner"] = exports.GlobalInterner;
// It's often useful to know just the sign of a multiplicity
function sign(x) {
    return typeof x === 'number' ? x ? x < 0 ? -1 : 1 : x === x ? 0 : NaN : NaN;
}
var Change = (function () {
    // Change expects that all values have already been interned.
    function Change(e, a, v, n, transaction, round, count) {
        this.e = e;
        this.a = a;
        this.v = v;
        this.n = n;
        this.transaction = transaction;
        this.round = round;
        this.count = count;
    }
    // As a convenience, you can generate a change from values that haven't been
    // interned yet.
    Change.fromValues = function (e, a, v, n, transaction, round, count) {
        return new Change(exports.GlobalInterner.intern(e), exports.GlobalInterner.intern(a), exports.GlobalInterner.intern(v), exports.GlobalInterner.intern(n), transaction, round, count);
    };
    Change.prototype.toString = function () {
        // let e = GlobalInterner.reverse(this.e);
        var e = this.e;
        return "Change(" + e + ", " + exports.GlobalInterner.reverse(this.a) + ", " + maybeReverse(this.v) + ", " + this.n + ", " + this.transaction + ", " + this.round + ", " + this.count + ")";
    };
    // For testing purposes, you often want to compare two Changes ignoring their
    // node, as you don't know exactly what node will generate a value when you
    // run. withoutE is also used in testing to check if a triple whose entity
    // may have been generated by the program *could* match this change.
    Change.prototype.equal = function (other, withoutNode, withoutE) {
        return (withoutE || this.e == other.e) &&
            this.a == other.a &&
            this.v == other.v &&
            (withoutNode || this.n == other.n) &&
            this.transaction == other.transaction &&
            this.round == other.round &&
            this.count == other.count;
    };
    Change.prototype.reverse = function (interner) {
        if (interner === void 0) { interner = exports.GlobalInterner; }
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n, transaction = _a.transaction, round = _a.round, count = _a.count;
        return new RawChange(interner.reverse(e), interner.reverse(a), interner.reverse(v), interner.reverse(n), transaction, round, count);
    };
    Change.prototype.toRawEAV = function (interner) {
        if (interner === void 0) { interner = exports.GlobalInterner; }
        var _a = this, e = _a.e, a = _a.a, v = _a.v;
        return [interner.reverse(e), interner.reverse(a), interner.reverse(v)];
    };
    Change.prototype.clone = function () {
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n, transaction = _a.transaction, round = _a.round, count = _a.count;
        return new Change(e, a, v, n, transaction, round, count);
    };
    return Change;
}());
exports.Change = Change;
var BLOCK_REMOVE = new Change(0, 0, 0, 0, 0, 0, -1);
var BLOCK_ADD = new Change(0, 0, 0, 0, 0, 0, 1);
var RemoveChange = (function (_super) {
    __extends(RemoveChange, _super);
    function RemoveChange() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    RemoveChange.prototype.toString = function () {
        // let e = GlobalInterner.reverse(this.e);
        var e = this.e;
        return "RemoveChange(" + e + ", " + exports.GlobalInterner.reverse(this.a) + ", " + maybeReverse(this.v) + ", " + this.n + ", " + this.transaction + ", " + this.round + ", " + this.count + ")";
    };
    RemoveChange.prototype.clone = function () {
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n, transaction = _a.transaction, round = _a.round, count = _a.count;
        return new RemoveChange(e, a, v, n, transaction, round, count);
    };
    return RemoveChange;
}(Change));
exports.RemoveChange = RemoveChange;
var RemoveVsChange = (function (_super) {
    __extends(RemoveVsChange, _super);
    function RemoveVsChange() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    RemoveVsChange.prototype.toRemoveChanges = function (context, changes) {
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n;
        var index = context.index, distinctIndex = context.distinctIndex;
        var matches = index.get(e, a, exports.IGNORE_REG, exports.IGNORE_REG, this.transaction, Infinity);
        for (var _i = 0, matches_1 = matches; _i < matches_1.length; _i++) {
            var v_1 = matches_1[_i].v;
            var rounds = index.getDiffs(e, a, v_1, exports.IGNORE_REG);
            for (var _b = 0, rounds_1 = rounds; _b < rounds_1.length; _b++) {
                var round = rounds_1[_b];
                var count = this.count * (round > 0 ? 1 : -1);
                var changeRound = Math.max(this.round, Math.abs(round) - 1);
                var change = new RemoveChange(e, a, v_1, n, this.transaction, changeRound, count);
                changes.push(change);
            }
        }
    };
    RemoveVsChange.prototype.clone = function () {
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n, transaction = _a.transaction, round = _a.round, count = _a.count;
        return new RemoveVsChange(e, a, v, n, transaction, round, count);
    };
    return RemoveVsChange;
}(RemoveChange));
exports.RemoveVsChange = RemoveVsChange;
var RemoveAVsChange = (function (_super) {
    __extends(RemoveAVsChange, _super);
    function RemoveAVsChange() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    RemoveAVsChange.prototype.toRemoveChanges = function (context, changes) {
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n;
        var index = context.index, distinctIndex = context.distinctIndex;
        var matches = index.get(e, exports.IGNORE_REG, exports.IGNORE_REG, exports.IGNORE_REG, this.transaction, Infinity);
        for (var _i = 0, matches_2 = matches; _i < matches_2.length; _i++) {
            var _b = matches_2[_i], a_1 = _b.a, v_2 = _b.v;
            var rounds = index.getDiffs(e, a_1, v_2, exports.IGNORE_REG);
            for (var _c = 0, rounds_2 = rounds; _c < rounds_2.length; _c++) {
                var round = rounds_2[_c];
                var count = this.count * (round > 0 ? 1 : -1);
                var changeRound = Math.max(this.round, Math.abs(round) - 1);
                var change = new RemoveChange(e, a_1, v_2, n, this.transaction, changeRound, count);
                changes.push(change);
            }
        }
    };
    RemoveAVsChange.prototype.clone = function () {
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n, transaction = _a.transaction, round = _a.round, count = _a.count;
        return new RemoveAVsChange(e, a, v, n, transaction, round, count);
    };
    return RemoveAVsChange;
}(RemoveVsChange));
exports.RemoveAVsChange = RemoveAVsChange;
// When interacting with the outside world, we need to pass changes around that
// are no longer interned. A RawChange is the same as Change, but all the
// information in the triple has been converted back into RawValues instead of
// interned IDs.
var RawChange = (function () {
    function RawChange(e, a, v, n, transaction, round, count) {
        this.e = e;
        this.a = a;
        this.v = v;
        this.n = n;
        this.transaction = transaction;
        this.round = round;
        this.count = count;
    }
    RawChange.prototype.toString = function () {
        var _a = this, e = _a.e, a = _a.a, v = _a.v, n = _a.n, transaction = _a.transaction, round = _a.round, count = _a.count;
        var internedE = exports.GlobalInterner.get(e);
        var internedV = exports.GlobalInterner.get(v);
        return "RawChange(" + internedE + ", " + a + ", " + (maybeReverse(internedV) || v) + ", " + n + ", " + transaction + ", " + round + ", " + count + ")";
    };
    return RawChange;
}());
exports.RawChange = RawChange;
// A register is a numerical offset into a prefix. We can't just make this a
// type alias to number because we need to be able to tell the difference between
// IDs which represent static values and registers which represent dynamic values
// in the prefix. For example I might have a constraint that looks for the
// pattern (register1, "tag", "person"), which if we treated Registers as numbers
// might just look like (1, 2, 3) after the values have been interned. Instead
// we make Register a class.
var Register = (function () {
    function Register(offset) {
        this.offset = offset;
    }
    return Register;
}());
exports.Register = Register;
function isRegister(x) {
    return x && x.constructor === Register;
}
exports.isRegister = isRegister;
// In some cases we have a constraint whose value we may want to ignore.
// IGNORE_REG is a sentinel value that tells us we don't care what the value of
// something is when we're solving.
exports.IGNORE_REG = null;
var ApplyInputState;
(function (ApplyInputState) {
    ApplyInputState[ApplyInputState["pass"] = 0] = "pass";
    ApplyInputState[ApplyInputState["fail"] = 1] = "fail";
    ApplyInputState[ApplyInputState["none"] = 2] = "none";
})(ApplyInputState = exports.ApplyInputState || (exports.ApplyInputState = {}));
var EAVN = (function () {
    function EAVN(e, a, v, n) {
        this.e = e;
        this.a = a;
        this.v = v;
        this.n = n;
    }
    return EAVN;
}());
exports.EAVN = EAVN;
;
//------------------------------------------------------------------------
// Move Constraint
//------------------------------------------------------------------------
var MoveConstraint = (function () {
    function MoveConstraint(from, to) {
        this.from = from;
        this.to = to;
        this.shouldApplyInput = false;
        this.proposal = { cardinality: 1, forFields: new Iterator(), forRegisters: new Iterator(), proposer: this };
        this.registers = createArray("MoveConstriantRegisters");
        this.resolved = createArray("MoveConstraintResolved");
        this.isInput = false;
        this.isStatic = true;
    }
    MoveConstraint.prototype.toString = function () {
        return "Move(" + printField(this.from) + ", " + printField(this.to) + ")";
    };
    MoveConstraint.prototype.setup = function () {
        if (isRegister(this.from)) {
            this.isStatic = false;
        }
        this.registers.push(this.to);
        // we are always only proposing for our to register
        this.proposal.forRegisters.clear();
        this.proposal.forRegisters.push(this.to);
    };
    MoveConstraint.prototype.resolve = function (prefix) {
        if (isRegister(this.from)) {
            this.resolved[0] = prefix[this.from.offset];
        }
        else {
            this.resolved[0] = this.from;
        }
        this.resolved[1] = prefix[this.to.offset];
        return this.resolved;
    };
    MoveConstraint.prototype.getRegisters = function () {
        return this.registers;
    };
    MoveConstraint.prototype.isAffected = function (input) {
        if (this.shouldApplyInput) {
            return ApplyInputState.pass;
        }
        return ApplyInputState.none;
    };
    MoveConstraint.prototype.applyInput = function (input, prefix) {
        if (this.shouldApplyInput) {
            var value = void 0;
            if (isRegister(this.from)) {
                value = prefix[this.from.offset];
            }
            else {
                value = this.from;
            }
            var current = prefix[this.to.offset];
            if (value !== undefined && (current === undefined || current == value)) {
                prefix[this.to.offset] = value;
            }
            else {
                return ApplyInputState.fail;
            }
            return ApplyInputState.pass;
        }
        return ApplyInputState.none;
    };
    MoveConstraint.prototype.propose = function (context, prefix, transaction, round, results) {
        var _a = this.resolve(prefix), from = _a[0], to = _a[1];
        this.proposal.skip = true;
        if (from !== undefined && to === undefined) {
            this.proposal.skip = false;
        }
        return this.proposal;
    };
    MoveConstraint.prototype.resolveProposal = function (context, prefix, proposal, transaction, round, results) {
        var _a = this.resolve(prefix), from = _a[0], to = _a[1];
        var arr = createArray("MoveResult");
        arr[0] = from;
        return arr;
    };
    MoveConstraint.prototype.accept = function (context, prefix, transaction, round, solvingFor) {
        var _a = this.resolve(prefix), from = _a[0], to = _a[1];
        if (from !== undefined && to !== undefined) {
            return from == to;
        }
        return true;
    };
    MoveConstraint.prototype.acceptInput = function (context, input, prefix, transaction, round) {
        return this.accept(context, prefix, transaction, round, this.registers);
    };
    MoveConstraint.prototype.getDiffs = function (context, prefix) {
        throw new Error("Asking for Diffs from MoveConstraint");
    };
    return MoveConstraint;
}());
exports.MoveConstraint = MoveConstraint;
//------------------------------------------------------------------------
// Scans
//------------------------------------------------------------------------
/**
 * A scan maps a set of bound variables to unbound variables.
 */
var Scan = (function () {
    function Scan(e, a, v, n) {
        this.e = e;
        this.a = a;
        this.v = v;
        this.n = n;
        this.resolved = { e: undefined, a: undefined, v: undefined, n: undefined };
        this.registers = createArray();
        this.registerLookup = createArray();
        this.isInput = false;
        this.proposal = { cardinality: 0, forFields: new Iterator(), forRegisters: new Iterator(), proposer: this };
    }
    Scan.prototype.toString = function () {
        return "Scan(" + printField(this.e) + " " + printField(this.a) + " " + printField(this.v) + " " + printField(this.n) + ")";
    };
    Scan.prototype.toKey = function () {
        var e = isRegister(this.e) ? "$reg(" + this.e.offset + ")" : this.e;
        var a = isRegister(this.a) ? "$reg(" + this.a.offset + ")" : this.a;
        var v = isRegister(this.v) ? "$reg(" + this.v.offset + ")" : this.v;
        return e + "|" + a + "|" + v;
    };
    /**
     * Resolve each scan field.
     * The resolved object may contain one of three possible value types:
     * - IGNORE_REG -- this field is entirely ignored by the scan.
     * - undefined -- this field is a register that hasn't been filled in yet.
     *                We'll fill it if possible.
     * - ID -- this field contains a static or already solved value.
     */
    Scan.prototype.resolve = function (prefix) {
        var resolved = this.resolved;
        if (isRegister(this.e)) {
            resolved.e = prefix[this.e.offset];
        }
        else {
            resolved.e = this.e;
        }
        if (isRegister(this.a)) {
            resolved.a = prefix[this.a.offset];
        }
        else {
            resolved.a = this.a;
        }
        if (isRegister(this.v)) {
            resolved.v = prefix[this.v.offset];
        }
        else {
            resolved.v = this.v;
        }
        if (isRegister(this.n)) {
            resolved.n = prefix[this.n.offset];
        }
        else {
            resolved.n = this.n;
        }
        return resolved;
    };
    /**
     * A field is unresolved if it is completely ignored by the scan or
     * is an output of the scan.
     */
    Scan.prototype.fieldUnresolved = function (resolved, key) {
        return resolved[key] === exports.IGNORE_REG || resolved[key] === undefined;
    };
    /**
     * A field is not a static match if it is ignored, not a static
     * field, or the input value does not match the static value.
     */
    Scan.prototype.notStaticMatch = function (input, key) {
        return this[key] !== exports.IGNORE_REG && !isRegister(this[key]) && this[key] !== input[key];
    };
    Scan.prototype.isAffected = function (input) {
        // If this change isn't relevant to this scan, skip it.
        if (this.notStaticMatch(input, "e"))
            return ApplyInputState.none;
        if (this.notStaticMatch(input, "a"))
            return ApplyInputState.none;
        if (this.notStaticMatch(input, "v"))
            return ApplyInputState.none;
        if (this.notStaticMatch(input, "n"))
            return ApplyInputState.none;
        return ApplyInputState.pass;
    };
    /**
     * Apply new changes that may affect this scan to the prefix to
     * derive only the results affected by this change.  If the change
     * was successfully applied or irrelevant we'll return true. If the
     * change was relevant but invalid (i.e., this scan could not be
     * satisfied due to proposals from previous scans) we'll return
     * false.
     */
    Scan.prototype.applyInput = function (input, prefix) {
        // For each register field of this scan:
        //   if the required value is impossible fail,
        //   else add this new value to the appropriate prefix register.
        // @NOTE: Technically, we republish existing values here too.
        //   In practice, that's harmless and eliminates the need for a branch.
        if (isRegister(this.e)) {
            if (prefix[this.e.offset] !== undefined && prefix[this.e.offset] !== input.e)
                return ApplyInputState.fail;
            prefix[this.e.offset] = input.e;
        }
        if (isRegister(this.a)) {
            if (prefix[this.a.offset] !== undefined && prefix[this.a.offset] !== input.a)
                return ApplyInputState.fail;
            prefix[this.a.offset] = input.a;
        }
        if (isRegister(this.v)) {
            if (prefix[this.v.offset] !== undefined && prefix[this.v.offset] !== input.v)
                return ApplyInputState.fail;
            prefix[this.v.offset] = input.v;
        }
        if (isRegister(this.n)) {
            if (prefix[this.n.offset] !== undefined && prefix[this.n.offset] !== input.n)
                return ApplyInputState.fail;
            prefix[this.n.offset] = input.n;
        }
        return ApplyInputState.pass;
    };
    Scan.prototype.propose = function (context, prefix, transaction, round, results) {
        var index = context.index;
        var _a = this.resolve(prefix), e = _a.e, a = _a.a, v = _a.v, n = _a.n;
        this.proposal.skip = false;
        var proposal = index.propose(this.proposal, e, a, v, n, transaction, round);
        var forRegisters = proposal.forRegisters, forFields = proposal.forFields;
        forRegisters.clear();
        var field;
        while ((field = forFields.next()) !== undefined) {
            forRegisters.push(this[field]);
        }
        if (proposal.forFields.length === 0)
            proposal.skip = true;
        return proposal;
    };
    Scan.prototype.resolveProposal = function (context, prefix, proposal, transaction, round, results) {
        var index = context.index;
        return index.resolveProposal(proposal);
    };
    Scan.prototype.accept = function (context, prefix, transaction, round, solvingFor) {
        // Before we start trying to accept, we check if we care about the
        // registers we are currently solving.
        var solving = false;
        for (var _i = 0, solvingFor_1 = solvingFor; _i < solvingFor_1.length; _i++) {
            var register = solvingFor_1[_i];
            if (this.registerLookup[register.offset]) {
                solving = true;
                break;
            }
        }
        // If we aren't looking at any of these registers, then we just
        // say we accept.
        if (!solving)
            return true;
        var _a = this.resolve(prefix), e = _a.e, a = _a.a, v = _a.v, n = _a.n;
        return context.index.check(e, a, v, n, transaction, round);
    };
    Scan.prototype.acceptInput = function (context, input, prefix, transaction, round) {
        var _a = this.resolve(prefix), e = _a.e, a = _a.a, v = _a.v, n = _a.n;
        if ((e === exports.IGNORE_REG || input.e === e) &&
            (a === exports.IGNORE_REG || input.a === a) &&
            (v === exports.IGNORE_REG || input.v === v) &&
            (n === exports.IGNORE_REG || input.n === n)) {
            return true;
        }
        else {
            return this.accept(context, prefix, transaction, round, this.registers);
        }
    };
    Scan.prototype.setupRegister = function (field, parts) {
        var value = this[field];
        if (isRegister(value)) {
            this.registers.push(value);
            parts.push("resolved." + field + " = prefix[" + value.offset + "];");
        }
        else {
            this.resolved[field] = value;
        }
    };
    Scan.prototype.setupIsAffected = function () {
        var fields = ["e", "a", "v", "n"];
        var parts = [];
        for (var _i = 0, fields_1 = fields; _i < fields_1.length; _i++) {
            var field = fields_1[_i];
            var value = this[field];
            if (!isRegister(value) && value !== exports.IGNORE_REG) {
                parts.push("if(" + value + " !== change[\"" + field + "\"]) return " + ApplyInputState.none + ";");
            }
        }
        this.isAffected = new Function("change", parts.join("\n"));
    };
    Scan.prototype.setupApplyInput = function () {
        var fields = ["e", "a", "v", "n"];
        var parts = [];
        for (var _i = 0, fields_2 = fields; _i < fields_2.length; _i++) {
            var field = fields_2[_i];
            var value = this[field];
            if (isRegister(value)) {
                parts.push("if(prefix[" + value.offset + "] !== undefined && prefix[" + value.offset + "] !== input." + field + ") return " + ApplyInputState.fail + ";\n                    prefix[" + value.offset + "] = input." + field + ";");
            }
        }
        parts.push("return " + ApplyInputState.pass);
        this.applyInput = new Function("input", "prefix", parts.join("\n"));
    };
    // We precompute the registers we're interested in for fast accepts.
    Scan.prototype.setup = function () {
        var parts = ["var resolved = this.resolved;"];
        this.setupRegister("e", parts);
        this.setupRegister("a", parts);
        this.setupRegister("v", parts);
        this.setupRegister("n", parts);
        parts.push("return resolved");
        this.resolve = new Function("prefix", parts.join("\n"));
        this.setupIsAffected();
        this.setupApplyInput();
        for (var _i = 0, _a = this.registers; _i < _a.length; _i++) {
            var register = _a[_i];
            this.registerLookup[register.offset] = true;
        }
    };
    Scan.prototype.getRegisters = function () {
        return this.registers;
    };
    Scan.prototype.getDiffs = function (context, prefix) {
        var _a = this.resolve(prefix), e = _a.e, a = _a.a, v = _a.v, n = _a.n;
        return context.index.getDiffs(e, a, v, n);
    };
    return Scan;
}());
exports.Scan = Scan;
var FunctionConstraint = (function () {
    function FunctionConstraint(fields, restFields) {
        this.fields = fields;
        this.restFields = restFields;
        this.multi = false;
        this.isInput = false;
        this.proposal = { cardinality: 0, forFields: new Iterator(), forRegisters: new Iterator(), proposer: this };
        this.resolved = {};
        this.resolvedRest = createArray();
        this.registers = createArray();
        this.registerLookup = createArray();
        this.applyInputs = createArray();
        this.applyRestInputs = createArray();
    }
    FunctionConstraint.register = function (name, klass) {
        FunctionConstraint.registered[name] = klass;
    };
    FunctionConstraint.fetchInfo = function (name) {
        var info = FunctionConstraint.registered[name];
        if (!info)
            throw new Error("No function info for: " + name);
        return info;
    };
    FunctionConstraint.create = function (name, fields, restFields) {
        if (restFields === void 0) { restFields = createArray(); }
        var cur = FunctionConstraint.registered[name];
        if (!cur) {
            throw new Error("No function named " + name + " is registered.");
        }
        if (restFields.length && !cur.variadic) {
            console.error("The " + name + " function is not variadic, so may not accept restFields.");
            restFields = createArray();
        }
        var created = new cur(fields, restFields);
        return created;
    };
    FunctionConstraint.prototype.toString = function () {
        var _this = this;
        var params = this.fieldNames.map(function (v) { return v + ": " + printField(_this.fields[v]); }).join(", ");
        var restParams = this.restFields.map(printField).join(", ");
        return "FunctionConstraint(\"" + this.name + "\", " + params + " " + (restParams ? ", [" + restParams + "]" : "") + ")";
    };
    // We precompute the registers we're interested in for fast accepts.
    FunctionConstraint.prototype.setup = function () {
        this.fieldNames = Object.keys(this.fields);
        for (var _i = 0, _a = this.fieldNames; _i < _a.length; _i++) {
            var fieldName = _a[_i];
            var field = this.fields[fieldName];
            if (isRegister(field))
                this.registers.push(field);
        }
        for (var _b = 0, _c = this.restFields; _b < _c.length; _b++) {
            var field = _c[_b];
            if (isRegister(field))
                this.registers.push(field);
        }
        for (var _d = 0, _e = this.registers; _d < _e.length; _d++) {
            var register = _e[_d];
            this.registerLookup[register.offset] = true;
        }
        this.setupResolve();
        this.setupResolveRest();
    };
    FunctionConstraint.prototype.setupResolve = function () {
        var resolved = this.resolved;
        var parts = ["var resolved = this.resolved;"];
        for (var _i = 0, _a = this.fieldNames; _i < _a.length; _i++) {
            var fieldName = _a[_i];
            var field = this.fields[fieldName];
            if (isRegister(field)) {
                parts.push("resolved[\"" + fieldName + "\"] = prefix[" + field.offset + "];");
            }
            else {
                resolved[fieldName] = field;
            }
        }
        parts.push("return resolved");
        this.resolve = new Function("prefix", parts.join("\n"));
    };
    FunctionConstraint.prototype.setupResolveRest = function () {
        var resolvedRest = this.resolvedRest;
        var parts = ["var resolvedRest = this.resolvedRest;"];
        var ix = 0;
        for (var _i = 0, _a = this.restFields; _i < _a.length; _i++) {
            var field = _a[_i];
            if (isRegister(field)) {
                parts.push("resolvedRest[" + ix + "] = prefix[" + field.offset + "]");
            }
            else {
                resolvedRest[ix] = field;
            }
            ix++;
        }
        parts.push("return resolvedRest;");
        this.resolveRest = new Function("prefix", parts.join("\n"));
    };
    FunctionConstraint.prototype.getRegisters = function () {
        return this.registers;
    };
    /**
     * Similar to `Scan.resolve`, but resolving a map of the function's
     * fields rather than an EAVN.
     */
    FunctionConstraint.prototype.resolve = function (prefix) {
        var resolved = this.resolved;
        for (var _i = 0, _a = this.fieldNames; _i < _a.length; _i++) {
            var fieldName = _a[_i];
            var field = this.fields[fieldName];
            if (isRegister(field)) {
                resolved[fieldName] = prefix[field.offset];
            }
            else {
                resolved[fieldName] = field;
            }
        }
        return resolved;
    };
    /**
     * If a function is variadic, we need to resolve its rest fields as well.
     */
    FunctionConstraint.prototype.resolveRest = function (prefix) {
        var resolvedRest = this.resolvedRest;
        var ix = 0;
        for (var _i = 0, _a = this.restFields; _i < _a.length; _i++) {
            var field = _a[_i];
            if (isRegister(field)) {
                resolvedRest[ix] = prefix[field.offset];
            }
            else {
                resolvedRest[ix] = field;
            }
            ix++;
        }
        return resolvedRest;
    };
    // Function constraints have nothing to apply to the input, so they
    // always return ApplyInputState.none
    FunctionConstraint.prototype.isAffected = function (input) { return ApplyInputState.none; };
    FunctionConstraint.prototype.applyInput = function (input, prefix) { return ApplyInputState.none; };
    FunctionConstraint.prototype.propose = function (context, prefix, transaction, round, results) {
        var proposal = this.proposal;
        proposal.forRegisters.clear();
        var resolved = this.resolve(prefix);
        // If none of our returns are unbound
        // @NOTE: We don't need to worry about the filter case here, since he'll be
        var unresolvedOutput = false;
        for (var _i = 0, _a = this.returnNames; _i < _a.length; _i++) {
            var output = _a[_i];
            if (resolved[output] === undefined) {
                unresolvedOutput = true;
                var field = this.fields[output];
                if (isRegister(field)) {
                    proposal.forRegisters.push(field);
                }
            }
        }
        if (!unresolvedOutput) {
            proposal.skip = true;
            return proposal;
        }
        // If any of our args aren't resolved yet, we can't compute results either.
        // @NOTE: This'll need to be touched up when we add optional support if they
        //   co-inhabit the args object.
        for (var _b = 0, _c = this.argNames; _b < _c.length; _b++) {
            var input = _c[_b];
            if (resolved[input] === undefined) {
                proposal.skip = true;
                return proposal;
            }
        }
        // Similarly, if we're variadic we need to check that all of our
        // variadic inputs bound to registers are resolved too.
        // @NOTE: We really need to bend over backwards at the moment to
        //   convince TS to check a static member of the current class...
        if (this.constructor.variadic) {
            var resolvedRest = this.resolveRest(prefix);
            for (var _d = 0, resolvedRest_1 = resolvedRest; _d < resolvedRest_1.length; _d++) {
                var field = resolvedRest_1[_d];
                if (field === undefined) {
                    proposal.skip = true;
                    return proposal;
                }
            }
        }
        // Otherwise, we're ready to propose.
        proposal.skip = false;
        if (this.estimate) {
            // If the function provides a cardinality estimator, invoke that.
            proposal.cardinality = this.estimate(context, prefix, transaction, round);
        }
        else {
            // Otherwise, we'll just return 1 for now, since computing a
            // function is almost always cheaper than a scan.
            // @NOTE: If this is an issue, we can just behave like scans and
            //   compute ourselves here, caching the results.
            proposal.cardinality = 1;
        }
        return proposal;
    };
    /**
     * Pack the resolved register values for the functions argument
     * fields into an array.
     */
    FunctionConstraint.prototype.packInputs = function (prefix) {
        var resolved = this.resolve(prefix);
        var inputs = this.applyInputs;
        var argIx = 0;
        for (var _i = 0, _a = this.argNames; _i < _a.length; _i++) {
            var argName = _a[_i];
            // If we're asked to resolve the propoal we know that we've
            // proposed, and we'll only propose if these are resolved.
            inputs[argIx] = exports.GlobalInterner.reverse(resolved[argName]);
            argIx++;
        }
        // If we're variadic, we also need to pack our var-args up and
        // attach them as the last argument.
        if (this.constructor.variadic) {
            var resolvedRest = this.resolveRest(prefix);
            var restInputs = this.applyRestInputs;
            restInputs.length = 0;
            var ix = 0;
            for (var _b = 0, resolvedRest_2 = resolvedRest; _b < resolvedRest_2.length; _b++) {
                var value = resolvedRest_2[_b];
                if (value !== undefined) {
                    restInputs[ix] = exports.GlobalInterner.reverse(value);
                }
                ix++;
            }
            inputs[argIx] = restInputs;
        }
        return inputs;
    };
    FunctionConstraint.prototype.unpackOutputs = function (outputs) {
        for (var ix = 0; ix < outputs.length; ix++) {
            // @NOTE: we'd like to use arenaIntern here, but because of intermediate values
            // that's not currently a possibility. We should revisit this if a practical solution
            // for arenas surfaces.
            outputs[ix] = exports.GlobalInterner.intern(outputs[ix]);
        }
        return outputs;
    };
    FunctionConstraint.prototype.getResult = function (prefix, outputs) {
        // Finally, if we had results, we create the result prefixes and pass them along.
        var result = createArray("functionResult");
        var ix = 0;
        for (var _i = 0, _a = this.returnNames; _i < _a.length; _i++) {
            var returnName = _a[_i];
            var field = this.fields[returnName];
            if (isRegister(field) && !prefix[field.offset]) {
                result[ix] = outputs[ix];
            }
            ix++;
        }
        return result;
    };
    FunctionConstraint.prototype.checkResult = function (prefix, outputs) {
        // Finally, we make sure every return register matches up with our outputs.
        // @NOTE: If we just use solvingFor then we don't know the offsets into the outputs array,
        // so we check everything...
        var ix = 0;
        for (var _i = 0, _a = this.returnNames; _i < _a.length; _i++) {
            var returnName = _a[_i];
            var field = this.fields[returnName];
            var value = isRegister(field) ? prefix[field.offset] : field;
            if (value !== outputs[ix]) {
                return false;
            }
            ix++;
        }
        return true;
    };
    FunctionConstraint.prototype.resolveProposal = function (context, prefix, proposal, transaction, round, results) {
        // First we build the args array to provide the apply function.
        var inputs = this.packInputs(prefix);
        // Then we actually apply it and then unpack the outputs.
        var computed = this.apply.apply(this, inputs);
        if (!computed)
            return results;
        if (!this.multi) {
            // If it's not a multi-returning function, it has a single result.
            var outputs = this.unpackOutputs(computed);
            var result = this.getResult(prefix, outputs);
            results.push(result);
        }
        else {
            for (var _i = 0, computed_1 = computed; _i < computed_1.length; _i++) {
                var row = computed_1[_i];
                // Otherwise it has N results.
                var outputs = this.unpackOutputs(row);
                var result = this.getResult(prefix, outputs);
                results.push(result);
            }
        }
        return results;
    };
    FunctionConstraint.prototype.accept = function (context, prefix, transaction, round, solvingFor) {
        // If none of the registers we're solving for intersect our inputs
        // or outputs, we're not relevant to the solution.
        var isRelevant = false;
        for (var _i = 0, solvingFor_2 = solvingFor; _i < solvingFor_2.length; _i++) {
            var register = solvingFor_2[_i];
            if (this.registerLookup[register.offset]) {
                isRelevant = true;
                break;
            }
        }
        if (!isRelevant)
            return true;
        // If we're missing a field, we can't verify our output yet so we preliminarily accept.
        for (var _a = 0, _b = this.fieldNames; _a < _b.length; _a++) {
            var fieldName = _b[_a];
            var field = this.fields[fieldName];
            if (isRegister(field) && prefix[field.offset] === undefined)
                return true;
        }
        // First we build the args array to provide the apply function.
        var inputs = this.packInputs(prefix);
        // Then we actually apply it and then unpack the outputs.
        var computed = this.apply.apply(this, inputs);
        if (!computed)
            return false;
        if (!this.multi) {
            // If it's not a multi-returning function we only need to check against the single result.
            var outputs = this.unpackOutputs(computed);
            return this.checkResult(prefix, outputs);
        }
        else {
            // Otherwise we match against any of the results.
            for (var _c = 0, computed_2 = computed; _c < computed_2.length; _c++) {
                var row = computed_2[_c];
                var outputs = this.unpackOutputs(row);
                if (this.checkResult(prefix, outputs))
                    return true;
            }
            return false;
        }
    };
    FunctionConstraint.prototype.acceptInput = function (context, input, prefix, transaction, round) {
        return this.accept(context, prefix, transaction, round, this.registers);
    };
    FunctionConstraint.prototype.getDiffs = function (context, prefix) {
        return [];
    };
    return FunctionConstraint;
}());
FunctionConstraint.registered = {};
FunctionConstraint.filter = false;
FunctionConstraint.variadic = false;
exports.FunctionConstraint = FunctionConstraint;
function _makeFunction(_a) {
    var name = _a.name, _b = _a.variadic, variadic = _b === void 0 ? false : _b, args = _a.args, returns = _a.returns, apply = _a.apply, estimate = _a.estimate, initialState = _a.initialState, _c = _a.multi, multi = _c === void 0 ? false : _c;
    var NewFunctionConstraint = (function (_super) {
        __extends(NewFunctionConstraint, _super);
        function NewFunctionConstraint() {
            var _this = _super !== null && _super.apply(this, arguments) || this;
            _this.name = name;
            _this.args = args;
            _this.argNames = Object.keys(args);
            _this.returnNames = Object.keys(returns);
            _this.returns = returns;
            _this.apply = apply;
            _this.state = initialState;
            _this.multi = multi;
            return _this;
        }
        return NewFunctionConstraint;
    }(FunctionConstraint));
    NewFunctionConstraint.variadic = variadic;
    NewFunctionConstraint.filter = Object.keys(returns).length === 0;
    NewFunctionConstraint.argNames = Object.keys(args);
    NewFunctionConstraint.returnNames = Object.keys(returns);
    FunctionConstraint.register(name, NewFunctionConstraint);
}
function makeFunction(args) {
    return _makeFunction(args);
}
exports.makeFunction = makeFunction;
function makeMultiFunction(args) {
    args.multi = true;
    return _makeFunction(args);
}
exports.makeMultiFunction = makeMultiFunction;
//------------------------------------------------------------------------
// Nodes
//------------------------------------------------------------------------
/**
 * Base class for nodes, the building blocks of blocks.
 */
var Node = (function () {
    function Node() {
        this.ID = Node.NodeID++;
        this.results = new Iterator();
    }
    Node.prototype.toBranchString = function () {
        return this.toString();
    };
    return Node;
}());
Node.NodeID = 0;
exports.Node = Node;
/**
 * The JoinNode implements generic join across multiple constraints.
 * Since our system is incremental, we need to do something slightly
 * fancier than we did in the previous runtime.  For each new change
 * that enters the system, we ask each of our constraints whether they
 * are capable of producing a new result. In the case where a single
 * constraint can, we presolve that constraint and then run the rest
 * normally, limited to only producing results that match the first
 * constraint. However, if multiple constraints might apply the input,
 * we need to run for each *combination* of heads. E.g.:
 *
 * Given a join node with constraints [A, B, C, and D], where A and D
 * can both apply the input, we must combine the results of the
 * following computations to get the full result set:
 *
 * Apply {A} -> Do {B, C, D}
 * Apply {A, D} -> Do {B, C}
 * Apply {D} -> Do {A, B, C}
 *
 * We calculate this using the power set in exec.
 *
 * We then apply each of these combinations by running a genericJoin
 * over the remaining unresolved registers.  We ask each un-applied
 * constraint to propose a register to be solved. If a constraint is
 * capable of resolving one, it returns the set of registers it can
 * resolve and an estimate of the result set's cardinality. Generic
 * Join chooses the cheapest proposal, which the winning constraint
 * then fully computes (or retrieves from cache and returns). Next it
 * asks each other constraint to accept or reject the proposal. If the
 * constraint doesn't apply to the solved registers, it accepts.  If
 * the solution contains results that match the output of the
 * constraint, it also accepts. Otherwise, it must reject the solution
 * and that particular run yields no results.
 */
var JoinNode = (function (_super) {
    __extends(JoinNode, _super);
    function JoinNode(constraints) {
        var _this = _super.call(this) || this;
        _this.constraints = constraints;
        _this.traceType = trace_1.TraceNode.Join;
        _this.isStatic = false;
        _this.dormant = false;
        _this.registerLength = 0;
        _this.emptyProposal = { cardinality: Infinity, forFields: new Iterator(), forRegisters: new Iterator(), skip: true, proposer: {} };
        _this.affectedConstraints = new Iterator();
        // We need to find all the registers contained in our scans so that
        // we know how many rounds of Generic Join we need to do.
        var registerLength = 0;
        var registerLookup = [];
        var registers = createArray();
        var proposedResultsArrays = createArray();
        var hasOnlyMoves = true;
        var hasNoScans = true;
        var onlyStatic = true;
        for (var _i = 0, constraints_1 = constraints; _i < constraints_1.length; _i++) {
            var constraint = constraints_1[_i];
            constraint.setup();
            if (!(constraint instanceof MoveConstraint))
                hasOnlyMoves = false;
            else if (!constraint.isStatic)
                onlyStatic = false;
            if (constraint instanceof Scan)
                hasNoScans = false;
            for (var _a = 0, _b = constraint.getRegisters(); _a < _b.length; _a++) {
                var register = _b[_a];
                if (!registerLookup[register.offset]) {
                    registers.push(createArray());
                    proposedResultsArrays.push(createArray());
                    registerLookup[register.offset] = true;
                    registerLength++;
                }
            }
        }
        if (hasOnlyMoves) {
            for (var _c = 0, _d = constraints; _c < _d.length; _c++) {
                var constraint = _d[_c];
                constraint.shouldApplyInput = true;
            }
            _this.isStatic = onlyStatic;
        }
        if (hasNoScans) {
            _this.exec = JoinNode.prototype.downStreamExec;
        }
        _this.registerLookup = registerLookup;
        _this.registerArrays = registers;
        _this.registerLength = registerLength;
        _this.proposedResultsArrays = proposedResultsArrays;
        return _this;
    }
    JoinNode.prototype.toString = function () {
        return "JoinNode([\n  " + this.constraints.map(toString).join("\n  ") + "\n])";
    };
    JoinNode.prototype.findAffectedConstraints = function (input, prefix) {
        // @TODO: Hoist me out.
        var affectedConstraints = this.affectedConstraints;
        affectedConstraints.clear();
        for (var ix = 0, len = this.constraints.length; ix < len; ix++) {
            var constraint = this.constraints[ix];
            var result = constraint.isAffected(input);
            if (result !== ApplyInputState.none) {
                affectedConstraints.push(constraint);
            }
        }
        return affectedConstraints;
    };
    JoinNode.prototype.applyCombination = function (context, input, prefix, transaction, round, results) {
        // debug("        Join combo:", prefix.slice());
        var countOfSolved = 0;
        for (var ix = 0; ix < this.registerLookup.length; ix++) {
            if (!this.registerLookup[ix])
                continue;
            if (prefix[ix] !== undefined)
                countOfSolved++;
        }
        var remainingToSolve = this.registerLength - countOfSolved;
        // context.tracer.tracker.blockTime("PresolveCheck");
        var valid = this.presolveCheck(context, input, prefix, transaction, round);
        // context.tracer.tracker.blockTimeEnd("PresolveCheck");
        // debug("        Join combo valid:", valid, remainingToSolve, countOfSolved, this.registerLength);
        if (!valid) {
            // do nothing
            return false;
        }
        else if (!remainingToSolve) {
            // if it is valid and there's nothing left to solve, then we've found
            // a full result and we should just continue
            this.prefixToResults(context, this.constraints, prefix, round, results);
            // debug("        Join combo result:", results);
            return true;
        }
        else {
            // debug("              GJ:", remainingToSolve, this.constraints);
            // For each node, find the new results that match the prefix.
            this.genericJoin(context, prefix, transaction, round, results, remainingToSolve);
            // context.tracer.tracker.blockTimeEnd("GenericJoin");
            return true;
        }
    };
    JoinNode.prototype.unapplyConstraint = function (constraint, prefix) {
        for (var _i = 0, _a = constraint.getRegisters(); _i < _a.length; _i++) {
            var register = _a[_i];
            prefix[register.offset] = undefined;
        }
    };
    JoinNode.prototype.presolveCheck = function (context, input, prefix, transaction, round) {
        var constraints = this.constraints;
        // window["counts"][window["active"]]++;
        for (var _i = 0, constraints_2 = constraints; _i < constraints_2.length; _i++) {
            var constraint = constraints_2[_i];
            var valid = constraint.acceptInput(context, input, prefix, transaction, round);
            if (!valid) {
                // debug("          INVALID:", constraint);
                return false;
            }
        }
        return true;
    };
    JoinNode.prototype.computeMultiplicities = function (context, results, prefix, currentRound, diffs, diffIndex) {
        if (diffIndex === void 0) { diffIndex = -1; }
        if (diffIndex === -1) {
            prefix[prefix.length - 2] = currentRound;
            prefix[prefix.length - 1] = this.inputCount;
            this.computeMultiplicities(context, results, prefix, currentRound, diffs, diffIndex + 1);
            prefix[prefix.length - 2] = undefined;
            prefix[prefix.length - 1] = undefined;
        }
        else if (diffIndex === diffs.length) {
            var result = copyArray(prefix, "gjResultsArray");
            context.tracer.capturePrefix(result);
            //debug("          GJ -> ", result);
            results.push(result);
        }
        else {
            var startingRound = prefix[prefix.length - 2];
            var startingMultiplicity = prefix[prefix.length - 1];
            var rounds = diffs[diffIndex];
            var roundToMultiplicity = {};
            var maxRound = currentRound;
            var ix = 0;
            var currentRoundCount = 0;
            for (var _i = 0, rounds_3 = rounds; _i < rounds_3.length; _i++) {
                var round = rounds_3[_i];
                if (Math.abs(round) - 1 > currentRound) {
                    break;
                }
                currentRoundCount += round > 0 ? 1 : -1;
                ix++;
            }
            if (currentRoundCount) {
                prefix[prefix.length - 2] = Math.max(currentRound, startingRound);
                prefix[prefix.length - 1] = startingMultiplicity * currentRoundCount;
                this.computeMultiplicities(context, results, prefix, currentRound, diffs, diffIndex + 1);
            }
            for (; ix < rounds.length; ix++) {
                var round = rounds[ix];
                var count = round > 0 ? 1 : -1;
                prefix[prefix.length - 2] = Math.max(Math.abs(round) - 1, startingRound);
                prefix[prefix.length - 1] = startingMultiplicity * count;
                this.computeMultiplicities(context, results, prefix, currentRound, diffs, diffIndex + 1);
            }
            prefix[prefix.length - 2] = startingRound;
            prefix[prefix.length - 1] = startingMultiplicity;
        }
        return results;
    };
    JoinNode.prototype.prefixToResults = function (context, constraints, prefix, round, results) {
        var diffs = [];
        for (var _i = 0, constraints_3 = constraints; _i < constraints_3.length; _i++) {
            var constraint = constraints_3[_i];
            if (constraint.isInput || !(constraint instanceof Scan))
                continue;
            var cur = constraint.getDiffs(context, prefix);
            diffs.push(cur);
        }
        this.computeMultiplicities(context, results, prefix, round, diffs);
    };
    JoinNode.prototype.genericJoin = function (context, prefix, transaction, round, results, roundIx) {
        if (roundIx === void 0) { roundIx = this.registerLength; }
        var _a = this, constraints = _a.constraints, emptyProposal = _a.emptyProposal;
        var proposedResults = this.proposedResultsArrays[roundIx - 1];
        var forRegisters = this.registerArrays[roundIx - 1];
        proposedResults.length = 0;
        var bestProposal = emptyProposal;
        for (var _i = 0, constraints_4 = constraints; _i < constraints_4.length; _i++) {
            var constraint = constraints_4[_i];
            var current = constraint.propose(context, prefix, transaction, round, proposedResults);
            if (!current.skip && current.cardinality === 0) {
                return results;
            }
            else if (current.cardinality < bestProposal.cardinality && !current.skip) {
                bestProposal = current;
            }
        }
        if (bestProposal.skip) {
            // debug("             BAILING", bestProposal);
            return results;
        }
        var proposer = bestProposal.proposer;
        // We have to copy here because we need to keep a reference to this even if later
        // rounds might overwrite the proposal
        moveArray(bestProposal.forRegisters.array, forRegisters);
        var resolved = proposer.resolveProposal(context, prefix, bestProposal, transaction, round, proposedResults);
        if (resolved[0] && resolved[0].constructor === Array) {
            resultLoop: for (var _b = 0, resolved_1 = resolved; _b < resolved_1.length; _b++) {
                var result = resolved_1[_b];
                var ix = 0;
                for (var _c = 0, forRegisters_1 = forRegisters; _c < forRegisters_1.length; _c++) {
                    var register = forRegisters_1[_c];
                    prefix[register.offset] = +result[ix];
                    ix++;
                }
                for (var _d = 0, constraints_5 = constraints; _d < constraints_5.length; _d++) {
                    var constraint = constraints_5[_d];
                    if (constraint === proposer)
                        continue;
                    if (!constraint.accept(context, prefix, transaction, round, forRegisters)) {
                        // debug("             BAILING", printConstraint(constraint));
                        continue resultLoop;
                    }
                }
                if (roundIx === 1) {
                    this.prefixToResults(context, constraints, prefix, round, results);
                }
                else {
                    this.genericJoin(context, prefix, transaction, round, results, roundIx - 1);
                }
            }
        }
        else {
            var register = forRegisters[0];
            resultLoop: for (var _e = 0, resolved_2 = resolved; _e < resolved_2.length; _e++) {
                var result = resolved_2[_e];
                prefix[register.offset] = +result;
                for (var _f = 0, constraints_6 = constraints; _f < constraints_6.length; _f++) {
                    var constraint = constraints_6[_f];
                    if (constraint === proposer)
                        continue;
                    if (!constraint.accept(context, prefix, transaction, round, forRegisters)) {
                        // debug("             BAILING", printConstraint(constraint));
                        continue resultLoop;
                    }
                }
                if (roundIx === 1) {
                    this.prefixToResults(context, constraints, prefix, round, results);
                }
                else {
                    this.genericJoin(context, prefix, transaction, round, results, roundIx - 1);
                }
            }
        }
        for (var _g = 0, forRegisters_2 = forRegisters; _g < forRegisters_2.length; _g++) {
            var register = forRegisters_2[_g];
            // @NOTE: marking this as any is spoopy at best, but since we should never
            // iterate over the prefix, but instead use it almost like a hash, this
            // should be fine.
            prefix[register.offset] = undefined;
        }
        return results;
    };
    JoinNode.prototype.downStreamExec = function (context, input, prefix, transaction, round, results) {
        if (this.isStatic && this.dormant)
            return false;
        this.inputCount = prefix[prefix.length - 1] !== undefined ? prefix[prefix.length - 1] : input.count;
        var inputRound = prefix[prefix.length - 2] !== undefined ? prefix[prefix.length - 2] : input.round;
        var didSomething = this.applyCombination(context, input, prefix, transaction, inputRound, results);
        if (this.isStatic && didSomething) {
            this.dormant = true;
        }
        return didSomething;
    };
    JoinNode.prototype.exec = function (context, input, prefix, transaction, round, results) {
        var didSomething = false;
        this.inputCount = input.count;
        // if we are removing the block, we need the join to compute all the results currently
        // being generated by this block and undo them. We do this by sending a blank prefix
        // through with an input count of -1.
        if (input === BLOCK_REMOVE) {
            didSomething = this.applyCombination(context, input, prefix, transaction, round, results);
        }
        else if (this.isStatic && this.dormant) {
            return false;
        }
        else if (input === BLOCK_ADD) {
            didSomething = this.applyCombination(context, input, prefix, transaction, round, results);
        }
        else {
            this.inputCount = input.count;
            var affectedConstraints = this.findAffectedConstraints(input, prefix);
            var combinationCount = Math.pow(2, affectedConstraints.length);
            for (var comboIx = combinationCount - 1; comboIx > 0; comboIx--) {
                //console.log("  Combo:", comboIx);
                var shouldApply = true;
                for (var constraintIx = 0; constraintIx < affectedConstraints.length; constraintIx++) {
                    var mask = 1 << constraintIx;
                    var isIncluded = (comboIx & mask) !== 0;
                    var constraint_1 = affectedConstraints.array[constraintIx];
                    constraint_1.isInput = isIncluded;
                    if (isIncluded) {
                        var valid = constraint_1.applyInput(input, prefix);
                        // debug("        included", printConstraint(constraint));
                        // If any member of the input constraints fails, this whole combination is doomed.
                        if (valid === ApplyInputState.fail) {
                            shouldApply = false;
                            break;
                        }
                        //console.log("    " + printConstraint(constraint));
                    }
                }
                //console.log("    ", printPrefix(prefix));
                if (shouldApply) {
                    didSomething = this.applyCombination(context, input, prefix, transaction, round, results) || didSomething;
                }
                var constraint_2 = void 0;
                affectedConstraints.reset();
                while ((constraint_2 = affectedConstraints.next()) !== undefined) {
                    this.unapplyConstraint(constraint_2, prefix);
                }
            }
            affectedConstraints.reset();
            var constraint = void 0;
            while ((constraint = affectedConstraints.next()) !== undefined) {
                constraint.isInput = false;
            }
        }
        if (this.isStatic && didSomething) {
            this.dormant = true;
        }
        return didSomething;
    };
    return JoinNode;
}(Node));
exports.JoinNode = JoinNode;
var DownstreamJoinNode = (function (_super) {
    __extends(DownstreamJoinNode, _super);
    function DownstreamJoinNode() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    DownstreamJoinNode.prototype.exec = function (context, input, prefix, transaction, round, results) {
        return this.downStreamExec(context, input, prefix, transaction, round, results);
    };
    return DownstreamJoinNode;
}(JoinNode));
exports.DownstreamJoinNode = DownstreamJoinNode;
var NoopJoinNode = (function (_super) {
    __extends(NoopJoinNode, _super);
    function NoopJoinNode() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    NoopJoinNode.prototype.exec = function (context, input, prefix, transaction, round, results) {
        if (!this.dormant) {
            this.prefixToResults(context, this.constraints, prefix, round, results);
            this.dormant = true;
            return true;
        }
        return false;
    };
    return NoopJoinNode;
}(JoinNode));
exports.NoopJoinNode = NoopJoinNode;
var WatchNode = (function (_super) {
    __extends(WatchNode, _super);
    function WatchNode(e, a, v, n, blockId) {
        var _this = _super.call(this) || this;
        _this.e = e;
        _this.a = a;
        _this.v = v;
        _this.n = n;
        _this.blockId = blockId;
        _this.traceType = trace_1.TraceNode.Watch;
        _this.resolved = {};
        _this.resolve = Scan.prototype.resolve;
        return _this;
    }
    WatchNode.prototype.toString = function () {
        return "WatchNode(" + printField(this.e) + ", " + printField(this.a) + ", " + printField(this.v) + ", " + printField(this.n) + ")";
    };
    WatchNode.prototype.exec = function (context, input, prefix, transactionId, round, results, transaction) {
        var resolved = this.resolve(prefix);
        var e = resolved.e, a = resolved.a, v = resolved.v, n = resolved.n;
        // @NOTE: This is wasteful.
        results.push(prefix);
        if (e === undefined || a === undefined || v === undefined || n === undefined) {
            throw new Error("Unable to produce an output with an undefined EAVN field [" + e + ", " + a + ", " + v + ", " + n + "]");
        }
        var prefixRound = prefix[prefix.length - 2];
        var prefixCount = prefix[prefix.length - 1];
        // @FIXME: Make sure I still work now that I'm sending all my deltas. I think I still need to use local intermediates.
        var change = new Change(e, a, v, n, transactionId, prefixRound + 1, prefixCount);
        transaction.export(context, this.blockId, change);
        return true;
    };
    return WatchNode;
}(Node));
exports.WatchNode = WatchNode;
var OutputWrapperNode = (function (_super) {
    __extends(OutputWrapperNode, _super);
    function OutputWrapperNode(nodes) {
        var _this = _super.call(this) || this;
        _this.nodes = nodes;
        _this.traceType = trace_1.TraceNode.Output;
        _this.binds = new Iterator();
        _this.commits = new Iterator();
        return _this;
    }
    OutputWrapperNode.prototype.toString = function () {
        return "OutputWrapper([" + (this.nodes.length ? "\n  " : "") + indent(this.nodes.map(toString).join("\n"), 2) + "])";
    };
    OutputWrapperNode.prototype.exec = function (context, input, prefix, transactionId, round, results, transaction) {
        var tracer = context.tracer;
        var _a = this, binds = _a.binds, commits = _a.commits;
        binds.clear();
        commits.clear();
        for (var _i = 0, _b = this.nodes; _i < _b.length; _i++) {
            var node = _b[_i];
            node.exec(context, input, prefix, transactionId, round, binds, commits);
        }
        binds.reset();
        var change;
        while (change = binds.next()) {
            transaction.output(context, change);
        }
        commits.reset();
        while (change = commits.next()) {
            transaction.commit(context, change);
        }
        return true;
    };
    return OutputWrapperNode;
}(Node));
exports.OutputWrapperNode = OutputWrapperNode;
var InsertNode = (function () {
    function InsertNode(e, a, v, n) {
        this.e = e;
        this.a = a;
        this.v = v;
        this.n = n;
        this.multiplier = 1;
        // We precompute the registers we're interested in for fast accepts.
        this.resolved = { e: undefined, a: undefined, v: undefined, n: undefined };
        var parts = ["var resolved = this.resolved;"];
        this.setupRegister("e", parts);
        this.setupRegister("a", parts);
        this.setupRegister("v", parts);
        this.setupRegister("n", parts);
        parts.push("return resolved");
        this.resolve = new Function("prefix", parts.join("\n"));
    }
    InsertNode.prototype.toString = function () {
        return "InsertNode(" + printField(this.e) + ", " + printField(this.a) + ", " + printField(this.v) + ", " + printField(this.n) + ")";
    };
    InsertNode.prototype.setupRegister = function (field, parts) {
        var value = this[field];
        if (isRegister(value)) {
            parts.push("resolved." + field + " = prefix[this." + field + ".offset];");
        }
        else {
            this.resolved[field] = value;
        }
    };
    InsertNode.prototype.output = function (change, binds, commits) {
        binds.push(change);
    };
    InsertNode.prototype.exec = function (context, input, prefix, transactionId, round, binds, commits) {
        var resolved = this.resolve(prefix);
        var e = resolved.e, a = resolved.a, v = resolved.v, n = resolved.n;
        if (e === undefined || a === undefined || v === undefined || n === undefined) {
            throw new Error("Unable to produce an output with an undefined EAVN field [" + e + ", " + a + ", " + v + ", " + n + "]");
        }
        var prefixRound = prefix[prefix.length - 2];
        var prefixCount = prefix[prefix.length - 1];
        var change = new Change(e, a, v, n, transactionId, prefixRound + 1, prefixCount * this.multiplier);
        this.output(change, binds, commits);
        return true;
    };
    return InsertNode;
}());
exports.InsertNode = InsertNode;
var CommitInsertNode = (function (_super) {
    __extends(CommitInsertNode, _super);
    function CommitInsertNode() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    CommitInsertNode.prototype.toString = function () {
        return "CommitInsertNode(" + printField(this.e) + ", " + printField(this.a) + ", " + printField(this.v) + ", " + printField(this.n) + ")";
    };
    CommitInsertNode.prototype.output = function (change, binds, commits) {
        commits.push(change);
    };
    return CommitInsertNode;
}(InsertNode));
exports.CommitInsertNode = CommitInsertNode;
var RemoveNode = (function (_super) {
    __extends(RemoveNode, _super);
    function RemoveNode() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.multiplier = -1;
        return _this;
    }
    RemoveNode.prototype.toString = function () {
        return "RemoveNode(" + printField(this.e) + ", " + printField(this.a) + ", " + printField(this.v) + ", " + printField(this.n) + ")";
    };
    RemoveNode.prototype.exec = function (context, input, prefix, transactionId, round, binds, commits) {
        var resolved = this.resolve(prefix);
        var e = resolved.e, a = resolved.a, v = resolved.v, n = resolved.n;
        if (e === undefined || a === undefined || (v === undefined && this.v !== exports.IGNORE_REG) || n === undefined) {
            return false;
        }
        var prefixRound = prefix[prefix.length - 2];
        var prefixCount = prefix[prefix.length - 1];
        if (this.v !== exports.IGNORE_REG) {
            var change = new RemoveChange(e, a, v, n, transactionId, prefixRound + 1, prefixCount * this.multiplier);
            this.output(change, binds, commits);
        }
        else if (this.a !== exports.IGNORE_REG) {
            var change = new RemoveVsChange(e, a, v, n, transactionId, prefixRound + 1, prefixCount * this.multiplier);
            this.output(change, binds, commits);
        }
        else {
            var change = new RemoveAVsChange(e, a, v, n, transactionId, prefixRound + 1, prefixCount * this.multiplier);
            this.output(change, binds, commits);
        }
        return true;
    };
    return RemoveNode;
}(InsertNode));
exports.RemoveNode = RemoveNode;
var CommitRemoveNode = (function (_super) {
    __extends(CommitRemoveNode, _super);
    function CommitRemoveNode() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.multiplier = -1;
        _this._exec = RemoveNode.prototype.exec;
        return _this;
    }
    CommitRemoveNode.prototype.toString = function () {
        return "CommitRemoveNode(" + printField(this.e) + ", " + printField(this.a) + ", " + printField(this.v) + ", " + printField(this.n) + ")";
    };
    CommitRemoveNode.prototype.exec = function (context, input, prefix, transactionId, round, binds, commits) {
        return this._exec(context, input, prefix, transactionId, round, binds, commits);
    };
    return CommitRemoveNode;
}(CommitInsertNode));
exports.CommitRemoveNode = CommitRemoveNode;
//------------------------------------------------------------------------------
// LinearFlow
//------------------------------------------------------------------------------
var LinearFlow = (function (_super) {
    __extends(LinearFlow, _super);
    function LinearFlow(nodes) {
        var _this = _super.call(this) || this;
        _this.nodes = nodes;
        _this.traceType = trace_1.TraceNode.LinearFlow;
        _this.results = new Iterator();
        _this.initialResults = new Iterator();
        return _this;
    }
    LinearFlow.prototype.toString = function () {
        var content = this.nodes.map(toString).join(",\n");
        return "LinearFlow([\n  " + indent(content, 2) + "\n])";
    };
    LinearFlow.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        this.initialResults.clear();
        this.initialResults.push(prefix);
        // We populate the prefix with values from the input change so we only derive the
        // results affected by it.
        var curPrefix;
        var iter = this.initialResults;
        for (var _i = 0, _a = this.nodes; _i < _a.length; _i++) {
            var node = _a[_i];
            if (iter.length === 0)
                return false;
            node.results.clear();
            while (curPrefix = iter.next()) {
                context.tracer.node(node, curPrefix);
                var valid = node.exec(context, input, curPrefix, transaction, round, node.results, changes);
                context.tracer.pop(trace_1.TraceFrameType.Node);
            }
            iter = node.results.iter();
        }
        while (curPrefix = iter.next()) {
            results.push(curPrefix);
        }
        return true;
    };
    return LinearFlow;
}(Node));
exports.LinearFlow = LinearFlow;
var IntermediateIndexIterator = (function () {
    function IntermediateIndexIterator() {
        this.valueIx = 0;
        this.countIx = 0;
        this.round = 0;
        this.count = 0;
        this.minRound = 0;
    }
    IntermediateIndexIterator.prototype.reset = function (values, minRound) {
        if (minRound === void 0) { minRound = 0; }
        this.values = values;
        this.valueIx = 0;
        this.countIx = 0;
        this.round = 0;
        this.count = 0;
        this.minRound = minRound + 1;
        this.valueKeys = Object.keys(values);
        this.currentCounts = values[this.valueKeys[0]];
        return this;
    };
    IntermediateIndexIterator.prototype.next = function () {
        var _a = this, currentCounts = _a.currentCounts, countIx = _a.countIx, minRound = _a.minRound;
        if (!currentCounts)
            return;
        countIx++;
        if (countIx >= currentCounts.length) {
            this.valueIx++;
            var nextValue = this.valueKeys[this.valueIx];
            currentCounts = this.currentCounts = this.values[nextValue];
            if (!currentCounts)
                return;
            countIx = 1;
        }
        var count = 0;
        if (countIx < minRound) {
            var total = 0;
            for (; countIx <= minRound; countIx++) {
                var cur = currentCounts[countIx];
                if (!cur)
                    continue;
                total += cur;
            }
            count = total;
            countIx = minRound;
        }
        else {
            for (; countIx < currentCounts.length; countIx++) {
                var cur = currentCounts[countIx];
                if (cur) {
                    count = cur;
                    break;
                }
            }
        }
        this.round = countIx - 1;
        this.count = count;
        this.countIx = countIx;
        if (count == 0)
            return this.next();
        return currentCounts[0];
    };
    return IntermediateIndexIterator;
}());
exports.IntermediateIndexIterator = IntermediateIndexIterator;
var IntermediateIndex = (function () {
    function IntermediateIndex() {
        this.index = {};
        this.iterator = new IntermediateIndexIterator();
    }
    IntermediateIndex.CreateKeyFunction = function (registers) {
        var items = registers.map(function (reg) {
            return "prefix[" + reg.offset + "]";
        });
        var code = "\n      return \"\" " + (items.length ? "+" : "") + " " + items.join(' + "|" + ') + ";\n      ";
        return new Function("prefix", code);
    };
    IntermediateIndex.prototype.insert = function (key, prefix) {
        var values = this.index[key];
        if (!values)
            values = this.index[key] = createHash("intermediateIndexValues");
        var valueKey = this.hashPrefix(prefix);
        var counts = values[valueKey];
        if (!counts) {
            counts = values[valueKey] = createArray("intermediateIndexCounts");
            counts[0] = prefix;
        }
        var round = prefix[prefix.length - 2] + 1;
        var count = prefix[prefix.length - 1];
        counts[round] = count + (counts[round] || 0);
        if (!counts[round]) {
            var shouldRemove = true;
            for (var ix = 1, len = counts.length; ix < len; ix++) {
                if (counts[ix]) {
                    shouldRemove = false;
                    break;
                }
            }
            if (shouldRemove) {
                delete values[valueKey];
            }
        }
    };
    IntermediateIndex.prototype.iter = function (key, round) {
        var values = this.index[key];
        if (values)
            return this.iterator.reset(values, round);
    };
    IntermediateIndex.prototype.hashPrefix = function (prefix) {
        var round = prefix[prefix.length - 2];
        var count = prefix[prefix.length - 1];
        prefix[prefix.length - 2] = undefined;
        prefix[prefix.length - 1] = undefined;
        var key = prefix.join("|");
        prefix[prefix.length - 2] = round;
        prefix[prefix.length - 1] = count;
        return key;
    };
    return IntermediateIndex;
}());
exports.IntermediateIndex = IntermediateIndex;
var ZeroingIterator = (function () {
    function ZeroingIterator() {
        this.roundIx = -1;
        this.countSum = 0;
        this.minRound = 0;
        this.count = 1;
    }
    ZeroingIterator.prototype.reset = function (counts, minRound) {
        if (minRound === void 0) { minRound = 0; }
        this.counts = counts;
        this.minRound = minRound;
        this.roundIx = -1;
        this.countSum = 0;
        this.count = 1;
        return this;
    };
    ZeroingIterator.prototype.next = function () {
        var _a = this, roundIx = _a.roundIx, counts = _a.counts, countSum = _a.countSum, minRound = _a.minRound;
        var countsLength = counts.length;
        roundIx++;
        if (roundIx >= countsLength)
            return;
        var final;
        if (roundIx <= minRound) {
            if (minRound >= countsLength)
                countsLength = minRound + 1;
            for (; roundIx < countsLength; roundIx++) {
                var cur = counts[roundIx];
                if (cur) {
                    countSum += cur;
                }
                if (roundIx >= minRound && countSum === 0) {
                    final = roundIx;
                    break;
                }
            }
        }
        else {
            for (; roundIx <= countsLength; roundIx++) {
                var cur = counts[roundIx];
                if (!cur)
                    continue;
                countSum += cur;
                if ((this.countSum === 0 && countSum > 0) ||
                    (this.countSum > 0 && countSum === 0)) {
                    final = roundIx;
                    break;
                }
            }
        }
        this.roundIx = roundIx;
        this.countSum = countSum;
        this.count = 1;
        if (countSum !== 0) {
            this.count = -1;
        }
        return final;
    };
    return ZeroingIterator;
}());
exports.ZeroingIterator = ZeroingIterator;
var KeyOnlyIntermediateIndex = (function () {
    function KeyOnlyIntermediateIndex() {
        this.index = {};
        this.iterator = new ZeroingIterator();
    }
    KeyOnlyIntermediateIndex.prototype.insert = function (key, prefix) {
        var counts = this.index[key];
        if (!counts)
            counts = this.index[key] = createArray("KeyOnlyIntermediateIndex");
        var round = prefix[prefix.length - 2];
        var count = prefix[prefix.length - 1];
        var prev = counts[round] || 0;
        counts[round] = count + prev;
        if (!counts[round]) {
            var shouldRemove = true;
            for (var ix = 0, len = counts.length; ix < len; ix++) {
                if (counts[ix]) {
                    shouldRemove = false;
                    break;
                }
            }
            if (shouldRemove) {
                delete this.index[key];
            }
        }
    };
    KeyOnlyIntermediateIndex.prototype.has = function (key) {
        return this.index[key] ? true : false;
    };
    KeyOnlyIntermediateIndex.prototype.iter = function (key, round) {
        var values = this.index[key];
        if (values)
            return this.iterator.reset(values, round);
    };
    return KeyOnlyIntermediateIndex;
}());
exports.KeyOnlyIntermediateIndex = KeyOnlyIntermediateIndex;
var BinaryFlow = (function (_super) {
    __extends(BinaryFlow, _super);
    function BinaryFlow(left, right) {
        var _this = _super.call(this) || this;
        _this.left = left;
        _this.right = right;
        _this.traceType = trace_1.TraceNode.BinaryJoin;
        return _this;
    }
    BinaryFlow.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        var _a = this, left = _a.left, right = _a.right;
        left.results.clear();
        context.tracer.node(left, prefix);
        left.exec(context, input, prefix, transaction, round, left.results, changes);
        context.tracer.pop(trace_1.TraceFrameType.Node);
        right.results.clear();
        context.tracer.node(right, prefix);
        right.exec(context, input, prefix, transaction, round, right.results, changes);
        context.tracer.pop(trace_1.TraceFrameType.Node);
        var result;
        var leftResults = left.results.iter();
        var rightResults = right.results.iter();
        while ((result = leftResults.next()) !== undefined) {
            this.onLeft(context, result, transaction, round, results);
        }
        while ((result = rightResults.next()) !== undefined) {
            this.onRight(context, result, transaction, round, results);
        }
        return true;
    };
    return BinaryFlow;
}(Node));
exports.BinaryFlow = BinaryFlow;
var BinaryJoinRight = (function (_super) {
    __extends(BinaryJoinRight, _super);
    function BinaryJoinRight(left, right, keyRegisters, registersToMerge) {
        var _this = _super.call(this, left, right) || this;
        _this.left = left;
        _this.right = right;
        _this.keyRegisters = keyRegisters;
        _this.registersToMerge = registersToMerge;
        _this.leftIndex = new IntermediateIndex();
        _this.rightIndex = new IntermediateIndex();
        _this._nodeName = "BinaryJoinRight";
        _this.keyFunc = IntermediateIndex.CreateKeyFunction(keyRegisters);
        return _this;
    }
    BinaryJoinRight.prototype.toString = function () {
        var keys = "[" + this.keyRegisters.map(function (r) { return "[" + r.offset + "]"; }).join(", ") + "]";
        var merge = "[" + this.registersToMerge.map(function (r) { return "[" + r.offset + "]"; }).join(", ") + "]";
        return this._nodeName + "({\n  keys: " + keys + ",\n  merge: " + merge + ",\n  left: " + indent(toString(this.left), 2) + ",\n  right: " + indent(toString(this.right), 2) + "\n})";
    };
    BinaryJoinRight.prototype.toBranchString = function () {
        var keys = "[" + this.keyRegisters.map(function (r) { return "[" + r.offset + "]"; }).join(", ") + "]";
        var merge = "[" + this.registersToMerge.map(function (r) { return "[" + r.offset + "]"; }).join(", ") + "]";
        return this._nodeName + "({\n  keys: " + keys + ",\n  merge: " + merge + ",\n  right: " + indent(toString(this.right), 2) + "\n})";
    };
    BinaryJoinRight.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        // debug("    Binary Join Right:")
        var _a = this, left = _a.left, right = _a.right;
        right.results.clear();
        context.tracer.node(right, prefix);
        right.exec(context, input, prefix, transaction, round, right.results, changes);
        context.tracer.pop(trace_1.TraceFrameType.Node);
        var leftResults = left.results.iter();
        var rightResults = right.results.iter();
        var result;
        while ((result = leftResults.next()) !== undefined) {
            this.onLeft(context, result, transaction, round, results);
        }
        while ((result = rightResults.next()) !== undefined) {
            this.onRight(context, result, transaction, round, results);
        }
        // debug("        results:", results.array.slice());
        return true;
    };
    BinaryJoinRight.prototype.merge = function (left, right) {
        for (var _i = 0, _a = this.registersToMerge; _i < _a.length; _i++) {
            var register = _a[_i];
            var leftValue = left[register.offset];
            var rightValue = right[register.offset];
            if (leftValue === undefined || leftValue === rightValue) {
                left[register.offset] = rightValue;
            }
            else {
                return false;
            }
        }
        return true;
    };
    BinaryJoinRight.prototype.onLeft = function (context, prefix, transaction, round, results) {
        var key = this.keyFunc(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var count = prefix[prefix.length - 1];
        this.leftIndex.insert(key, prefix);
        var diffs = this.rightIndex.iter(key, round);
        // debug("       join left", key, printPrefix(prefix), diffs);
        if (!diffs)
            return;
        var rightPrefix;
        while (rightPrefix = diffs.next()) {
            var result = copyArray(prefix, "BinaryJoinResult");
            if (this.merge(result, rightPrefix)) {
                result[result.length - 2] = Math.max(prefixRound, diffs.round);
                result[result.length - 1] = count * diffs.count;
                context.tracer.capturePrefix(result);
                results.push(result);
                // debug("               join left -> ", printPrefix(result), diffs.round, count, diffs.count);
            }
        }
    };
    BinaryJoinRight.prototype.onRight = function (context, prefix, transaction, round, results) {
        var key = this.keyFunc(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var count = prefix[prefix.length - 1];
        this.rightIndex.insert(key, prefix);
        var diffs = this.leftIndex.iter(key, round);
        // debug("       join right", key, this.rightIndex.index[key]);
        if (!diffs)
            return;
        var leftPrefix;
        while (leftPrefix = diffs.next()) {
            var result = copyArray(leftPrefix, "BinaryJoinResult");
            if (this.merge(result, prefix)) {
                result[result.length - 2] = Math.max(prefixRound, diffs.round);
                result[result.length - 1] = count * diffs.count;
                context.tracer.capturePrefix(result);
                results.push(result);
                // debug("              join right -> ", printPrefix(result.slice()), diffs.round, count, diffs.count);
            }
        }
    };
    return BinaryJoinRight;
}(BinaryFlow));
exports.BinaryJoinRight = BinaryJoinRight;
var AntiJoin = (function (_super) {
    __extends(AntiJoin, _super);
    function AntiJoin(left, right, keyRegisters) {
        var _this = _super.call(this, left, right) || this;
        _this.left = left;
        _this.right = right;
        _this.keyRegisters = keyRegisters;
        _this.traceType = trace_1.TraceNode.AntiJoin;
        _this.leftIndex = new IntermediateIndex();
        _this.rightIndex = new KeyOnlyIntermediateIndex();
        _this.distinct = new indexes_1.DistinctIndex();
        _this.keyFunc = IntermediateIndex.CreateKeyFunction(keyRegisters);
        return _this;
    }
    AntiJoin.prototype.toString = function () {
        var left = indent(toString(this.left), 2);
        var right = indent(toString(this.right), 2);
        return "Antijoin({\n  left: " + left + ",\n  right: " + right + "\n})";
    };
    AntiJoin.prototype.toBranchString = function () {
        var right = indent(toString(this.right), 2);
        return "Antijoin({\n  right: " + right + "\n})";
    };
    AntiJoin.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        // debug("            antijoin:")
        return _super.prototype.exec.call(this, context, input, prefix, transaction, round, results, changes);
    };
    AntiJoin.prototype.onLeft = function (context, prefix, transaction, round, results) {
        var key = this.keyFunc(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var count = prefix[prefix.length - 1];
        this.leftIndex.insert(key, prefix);
        var diffs = this.rightIndex.iter(key, prefixRound);
        // debug("                left:", key, count, this.rightIndex.index[key] && copyArray(this.rightIndex.index[key]), prefix);
        if (!diffs) {
            // debug("                    left ->", key, count, diffs)
            return results.push(prefix);
        }
        else {
            var currentRound = void 0;
            while ((currentRound = diffs.next()) !== undefined) {
                var result = copyArray(prefix, "AntiJoinResult");
                result[result.length - 2] = currentRound;
                result[result.length - 1] = diffs.count * count;
                context.tracer.capturePrefix(result);
                results.push(result);
                // debug("                    left ->", key, count, currentRound, result[result.length - 1])
            }
        }
    };
    AntiJoin.prototype.onRight = function (context, prefix, transaction, round, results) {
        var key = this.keyFunc(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var count = prefix[prefix.length - 1];
        this.rightIndex.insert(key, prefix);
        var neue = new Iterator();
        this.distinct.distinctKey(key, prefixRound, count, neue);
        var diffs = this.leftIndex.iter(key, prefixRound);
        var copy = function (thing) {
            var neue = copyHash(thing);
            for (var key_1 in thing) {
                neue[key_1] = thing[key_1].slice();
            }
            return neue;
        };
        // debug("                right:", key, count, this.leftIndex.index[key] && copy(this.leftIndex.index[key]));
        // debug("                right distinct: ", this.distinct.index[key], neue);
        if (!diffs || !neue.length)
            return;
        var leftPrefix;
        var rightDelta;
        while (rightDelta = neue.next()) {
            var rightRound = rightDelta[0], rightCount = rightDelta[1];
            diffs = this.leftIndex.iter(key, prefixRound); // We already checked for this above.
            while (leftPrefix = diffs.next()) {
                var result = copyArray(leftPrefix, "AntiJoinResult");
                var maxRound = Math.max(diffs.round, rightRound);
                result[result.length - 2] = maxRound;
                result[result.length - 1] = rightCount * diffs.count * -1;
                context.tracer.capturePrefix(result);
                results.push(result);
                // debug("                    right ->", key, maxRound, rightCount, diffs.count, result[result.length - 1])
            }
        }
    };
    return AntiJoin;
}(BinaryFlow));
exports.AntiJoin = AntiJoin;
var AntiJoinPresolvedRight = (function (_super) {
    __extends(AntiJoinPresolvedRight, _super);
    function AntiJoinPresolvedRight() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this.traceType = trace_1.TraceNode.AntiJoinPresolvedRight;
        return _this;
    }
    AntiJoinPresolvedRight.prototype.toString = function () {
        return "AntiJoinPresolvedRight(" + toString(this.left) + ")";
    };
    AntiJoinPresolvedRight.prototype.toBranchString = function () {
        return "AntiJoinPresolvedRight(" + this.left.toBranchString() + ")";
    };
    AntiJoinPresolvedRight.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        var _a = this, left = _a.left, right = _a.right;
        left.results.clear();
        context.tracer.node(left, prefix);
        left.exec(context, input, prefix, transaction, round, left.results, changes);
        context.tracer.pop(trace_1.TraceFrameType.Node);
        var leftResults = left.results.iter();
        var rightResults = right.results.iter();
        var result;
        while ((result = leftResults.next()) !== undefined) {
            this.onLeft(context, result, transaction, round, results);
        }
        while ((result = rightResults.next()) !== undefined) {
            this.onRight(context, result, transaction, round, results);
        }
        return true;
    };
    return AntiJoinPresolvedRight;
}(AntiJoin));
exports.AntiJoinPresolvedRight = AntiJoinPresolvedRight;
var UnionFlow = (function (_super) {
    __extends(UnionFlow, _super);
    function UnionFlow(left, branches, keyRegisters, registersToMerge, extraOuterJoins) {
        var _this = _super.call(this) || this;
        _this.left = left;
        _this.keyRegisters = keyRegisters;
        _this.registersToMerge = registersToMerge;
        _this.extraOuterJoins = extraOuterJoins;
        _this.traceType = trace_1.TraceNode.Union;
        _this.branches = [];
        _this.emptyResults = new Iterator();
        var ix = 0;
        for (var _i = 0, branches_1 = branches; _i < branches_1.length; _i++) {
            var branch = branches_1[_i];
            _this.branches.push(new BinaryJoinRight(left, branch, keyRegisters[ix].concat(extraOuterJoins), registersToMerge));
            ix++;
        }
        return _this;
    }
    UnionFlow.prototype.toString = function () {
        var name;
        var branchText = this.branches.map(function (branch) { return indent(branch.toBranchString(), 4); }).join(",\n    ");
        return "UnionFlow({\n  left: " + indent(toString(this.left), 2) + ",\n  branches: [\n    " + branchText + "\n)}]\n})";
    };
    UnionFlow.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        var left = this.left;
        var tracer = context.tracer;
        var tempLeftResults = left.results;
        left.results.clear();
        tracer.node(left, prefix);
        left.exec(context, input, prefix, transaction, round, left.results, changes);
        tracer.pop(trace_1.TraceFrameType.Node);
        var leftPrefix;
        var leftResults = left.results.iter();
        for (var _i = 0, _a = this.branches; _i < _a.length; _i++) {
            var node = _a[_i];
            node.results.clear();
            tracer.node(node, prefix);
            node.exec(context, input, prefix, transaction, round, node.results, changes);
            tracer.pop(trace_1.TraceFrameType.Node);
            // Because we've already run this node once, we don't want it to potentially see the left's
            // results multiple times. As such, we temporarily set the results to an empty iterator
            // so that downstream nodes see nothing and we set it back once we've gone through
            // all the left prefixes.
            left.results = this.emptyResults;
            leftResults.reset();
            if (node.keyRegisters.length && input !== BLOCK_ADD && input !== BLOCK_REMOVE) {
                while ((leftPrefix = leftResults.next()) !== undefined) {
                    tracer.node(node, leftPrefix);
                    node.exec(context, input, copyArray(leftPrefix, "UnionLeftPrefixCopy"), transaction, round, node.results, changes);
                    tracer.pop(trace_1.TraceFrameType.Node);
                }
            }
            // set the left results back to the real results
            left.results = tempLeftResults;
            var branchResults = node.results.iter();
            var result = void 0;
            while ((result = branchResults.next())) {
                results.push(result);
            }
        }
        return true;
    };
    return UnionFlow;
}(Node));
exports.UnionFlow = UnionFlow;
var ChooseFlow = (function (_super) {
    __extends(ChooseFlow, _super);
    function ChooseFlow(left, initialBranches, keyRegisters, registersToMerge, extraOuterJoins) {
        var _this = _super.call(this) || this;
        _this.left = left;
        _this.keyRegisters = keyRegisters;
        _this.registersToMerge = registersToMerge;
        _this.extraOuterJoins = extraOuterJoins;
        _this.traceType = trace_1.TraceNode.Choose;
        _this.leftResults = new Iterator();
        _this.emptyResults = new Iterator();
        _this.branches = [];
        var allKeys = [];
        for (var _i = 0, keyRegisters_1 = keyRegisters; _i < keyRegisters_1.length; _i++) {
            var keySet = keyRegisters_1[_i];
            var _loop_1 = function (key) {
                if (!allKeys.some(function (r) { return r.offset === key.offset; })) {
                    allKeys.push(key);
                }
            };
            for (var _a = 0, keySet_1 = keySet; _a < keySet_1.length; _a++) {
                var key = keySet_1[_a];
                _loop_1(key);
            }
        }
        var branches = _this.branches;
        var prev;
        var ix = 0;
        for (var _b = 0, initialBranches_1 = initialBranches; _b < initialBranches_1.length; _b++) {
            var branch = initialBranches_1[_b];
            var myKeys = keyRegisters[ix].concat(extraOuterJoins);
            var join = void 0;
            if (prev) {
                join = new BinaryJoinRight(left, branch, myKeys, registersToMerge);
                var antijoin = new AntiJoinPresolvedRight(join, _this, allKeys);
                branches.push(antijoin);
            }
            else {
                join = new BinaryJoinRight(left, branch, myKeys, registersToMerge);
                branches.push(join);
            }
            prev = join;
            ix++;
        }
        return _this;
    }
    ChooseFlow.prototype.toString = function () {
        var name;
        var branchText = this.branches.map(function (branch) { return indent(branch.toBranchString(), 4); }).join(",\n    ");
        return "ChooseFlow({\n  left: " + indent(toString(this.left), 2) + ",\n  branches: [\n    " + branchText + "\n)}]\n})";
    };
    ChooseFlow.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        var tracer = context.tracer;
        var _a = this, branches = _a.branches, left = _a.left;
        var prev;
        var ix = 0;
        var tempResults = this.results;
        var tempLeftResults = left.results;
        left.results.clear();
        tracer.node(left, prefix);
        left.exec(context, input, prefix, transaction, round, left.results, changes);
        tracer.pop(trace_1.TraceFrameType.Node);
        var leftResults = left.results.iter();
        var leftPrefix;
        for (var _i = 0, branches_2 = branches; _i < branches_2.length; _i++) {
            var node = branches_2[_i];
            node.results.clear();
            tracer.node(node, prefix);
            node.exec(context, input, prefix, transaction, round, node.results, changes);
            tracer.pop(trace_1.TraceFrameType.Node);
            // Because we've already run this node once, we don't want it to potentially see our
            // results multiple times. As such, we temporarily set our results to an empty iterator
            // so that downstream nodes see nothing and we set it back once we've gone through
            // all the left prefixes. This ensures that AntiJoinPresolvedRight only sees the previous
            // branches' results once. We also need to do this for our left's results, since we'll have
            // seen those as well and would otherwise double count them.
            this.results = this.emptyResults;
            left.results = this.emptyResults;
            leftResults.reset();
            if (node.keyRegisters.length && input !== BLOCK_ADD && input !== BLOCK_REMOVE) {
                while ((leftPrefix = leftResults.next()) !== undefined) {
                    tracer.node(node, leftPrefix);
                    node.exec(context, input, copyArray(leftPrefix, "ChooseLeftPrefixCopy"), transaction, round, node.results, changes);
                    tracer.pop(trace_1.TraceFrameType.Node);
                }
            }
            // per above, make sure we set our results back to the real iterator
            this.results = tempResults;
            left.results = tempLeftResults;
            var branchResult = node.results.iter();
            var result = void 0;
            while ((result = branchResult.next()) !== undefined) {
                tracer.capturePrefix(result);
                results.push(result);
            }
        }
        return true;
    };
    return ChooseFlow;
}(Node));
exports.ChooseFlow = ChooseFlow;
var MergeAggregateFlow = (function (_super) {
    __extends(MergeAggregateFlow, _super);
    function MergeAggregateFlow() {
        var _this = _super !== null && _super.apply(this, arguments) || this;
        _this._nodeName = "MergeAggregateFlow";
        return _this;
    }
    MergeAggregateFlow.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        // debug("        AGG MERGE");
        var result;
        var _a = this, left = _a.left, right = _a.right;
        left.results.clear();
        left.exec(context, input, prefix, transaction, round, left.results, changes);
        // debug("              left results: ", leftResults);
        var leftResults = left.results.iter();
        // we run the left's results through the aggregate to capture all the aggregate updates
        right.results.clear();
        while ((result = leftResults.next()) !== undefined) {
            // debug("              left result: ", result.slice());
            right.exec(context, input, result, transaction, round, right.results, changes);
        }
        // now we go through all the lefts and rights like normal
        leftResults.reset();
        while ((result = leftResults.next()) !== undefined) {
            this.onLeft(context, result, transaction, round, results);
        }
        var rightResults = right.results.iter();
        while ((result = rightResults.next()) !== undefined) {
            this.onRight(context, result, transaction, round, results);
        }
        return true;
    };
    return MergeAggregateFlow;
}(BinaryJoinRight));
exports.MergeAggregateFlow = MergeAggregateFlow;
// This node is a bit strange, but is required to make sure that aggregates
// that are inside of a choose don't end up seeing results that wouldn't actually
// join with the outer scope of the choose. For example if we have the following rule:
//
// prog.block("count the names of people", ({find, gather, record, choose}) => {
//   let person = find("person");
//   let [sort] = choose(() => {
//     return gather(person.name).count();
//   }, () => "yo yo yo");
//   return [person.add("next", sort)];
// });
//
// If we join the choose branch to the outer *after* we've aggregated, then we're
// going to count everything with a name whether they're a person or not. Instead
// we need to make sure there is a join with the outer scope before it makes it to
// the choose. To do that, the AggregateOuterLookup node just keeps track of every
// value it has seen from the outer and makes sure that each right has a join with
// it.
var AggregateOuterLookup = (function (_super) {
    __extends(AggregateOuterLookup, _super);
    function AggregateOuterLookup(left, right, keyRegisters) {
        var _this = _super.call(this, left, right) || this;
        _this.left = left;
        _this.right = right;
        _this.keyRegisters = keyRegisters;
        _this._nodeName = "AggregateOuterLookup";
        _this.traceType = trace_1.TraceNode.AggregateOuterLookup;
        _this.leftIndex = new KeyOnlyIntermediateIndex();
        _this.rightIndex = new IntermediateIndex();
        _this.keyFunc = IntermediateIndex.CreateKeyFunction(keyRegisters);
        return _this;
    }
    AggregateOuterLookup.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        var _a = this, left = _a.left, right = _a.right;
        right.results.clear();
        context.tracer.node(right, prefix);
        right.exec(context, input, prefix, transaction, round, right.results, changes);
        context.tracer.pop(trace_1.TraceFrameType.Node);
        var result;
        var leftResults = left.results.iter();
        while ((result = leftResults.next()) !== undefined) {
            this.onLeft(context, result, transaction, round, results);
        }
        var rightResults = right.results.iter();
        while ((result = rightResults.next()) !== undefined) {
            this.onRight(context, result, transaction, round, results);
        }
        return true;
    };
    AggregateOuterLookup.prototype.onLeft = function (context, prefix, transaction, round, results) {
        var key = this.keyFunc(prefix);
        var exists = this.leftIndex.has(key);
        this.leftIndex.insert(key, prefix);
        var afterExists = this.leftIndex.has(key);
        var diffs = this.rightIndex.iter(key, 0);
        if (exists && afterExists || !diffs)
            return;
        var multiplier = 1;
        if (exists && !afterExists) {
            // remove
            multiplier = -1;
        }
        var rightPrefix;
        while (rightPrefix = diffs.next()) {
            var result = copyArray(rightPrefix, "aggregateLookupResult");
            result[result.length - 2] = diffs.round;
            result[result.length - 1] = diffs.count * multiplier;
            context.tracer.capturePrefix(result);
            results.push(result);
        }
    };
    AggregateOuterLookup.prototype.onRight = function (context, prefix, transaction, round, results) {
        var key = this.keyFunc(prefix);
        this.rightIndex.insert(key, prefix);
        if (this.leftIndex.has(key)) {
            results.push(prefix);
        }
    };
    return AggregateOuterLookup;
}(BinaryFlow));
exports.AggregateOuterLookup = AggregateOuterLookup;
var AggregateNode = (function (_super) {
    __extends(AggregateNode, _super);
    // @TODO: allow for multiple returns
    function AggregateNode(groupRegisters, projectRegisters, inputs, resultRegisters) {
        var _this = _super.call(this) || this;
        _this.groupRegisters = groupRegisters;
        _this.projectRegisters = projectRegisters;
        _this.inputs = inputs;
        _this.resultRegisters = resultRegisters;
        _this.traceType = trace_1.TraceNode.Aggregate;
        _this.groups = {};
        _this.resolved = [];
        _this.registerLookup = [];
        _this.groupKey = IntermediateIndex.CreateKeyFunction(groupRegisters);
        _this.projectKey = IntermediateIndex.CreateKeyFunction(projectRegisters);
        for (var _i = 0, groupRegisters_1 = groupRegisters; _i < groupRegisters_1.length; _i++) {
            var reg = groupRegisters_1[_i];
            _this.registerLookup[reg.offset] = true;
        }
        for (var _a = 0, resultRegisters_1 = resultRegisters; _a < resultRegisters_1.length; _a++) {
            var reg = resultRegisters_1[_a];
            _this.registerLookup[reg.offset] = true;
        }
        return _this;
    }
    AggregateNode.prototype.toString = function () {
        var groups = printFieldArray(this.groupRegisters);
        var projects = printFieldArray(this.projectRegisters);
        var inputs = printFieldArray(this.inputs);
        var results = printFieldArray(this.resultRegisters);
        return "AggregateNode(" + this.name + ", " + groups + ", " + projects + ", " + inputs + ", " + results + ")";
    };
    AggregateNode.prototype.groupPrefix = function (group, prefix) {
        var projection = this.projectKey(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var prefixCount = prefix[prefix.length - 1];
        var delta = 0;
        var found = this.groups[group];
        if (!found) {
            found = this.groups[group] = { result: [] };
        }
        var counts = found[projection] || [];
        var totalCount = 0;
        var countIx = 0;
        for (var _i = 0, counts_1 = counts; _i < counts_1.length; _i++) {
            var count = counts_1[_i];
            // we need the total up to our current round
            if (countIx > prefixRound)
                break;
            countIx++;
            if (!count)
                continue;
            totalCount += count;
        }
        if (totalCount && totalCount + prefixCount <= 0) {
            // subtract
            delta = -1;
        }
        else if (totalCount === 0 && totalCount + prefixCount > 0) {
            // add
            delta = 1;
        }
        else if (totalCount + prefixCount < 0) {
            // we have removed more values than exist?
            throw new Error("Negative total count for an aggregate projection");
        }
        else {
            // otherwise this change doesn't impact the projected count, we've just added
            // or removed a support.
        }
        counts[prefixRound] = (counts[prefixRound] || 0) + prefixCount;
        found[projection] = counts;
        return delta;
    };
    AggregateNode.prototype.getResultPrefix = function (prefix, result, count) {
        var neue = copyArray(prefix, "aggregateResult");
        neue[this.resultRegisters[0].offset] = result;
        neue[neue.length - 1] = count;
        var ix = 0;
        while (ix < neue.length - 2) {
            if (!this.registerLookup[ix]) {
                neue[ix] = undefined;
            }
            ix++;
        }
        return neue;
    };
    AggregateNode.prototype.resolve = function (prefix) {
        var resolved = this.resolved;
        var ix = 0;
        for (var _i = 0, _a = this.inputs; _i < _a.length; _i++) {
            var field = _a[_i];
            if (isRegister(field)) {
                resolved[ix] = exports.GlobalInterner.reverse(prefix[field.offset]);
            }
            else {
                resolved[ix] = exports.GlobalInterner.reverse(field);
            }
            ix++;
        }
        return resolved;
    };
    AggregateNode.prototype.stateToResult = function (state) {
        var current = this.getResult(state);
        return exports.GlobalInterner.intern(current);
    };
    AggregateNode.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        var group = this.groupKey(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var delta = this.groupPrefix(group, prefix);
        var op = this.add;
        if (!delta)
            return false;
        if (delta < 0)
            op = this.remove;
        var groupStates = this.groups[group].result;
        var currentState = groupStates[prefixRound];
        if (!currentState) {
            // otherwise we have to find the most recent result that we've seen
            for (var ix = 0, len = Math.min(groupStates.length, prefixRound); ix < len; ix++) {
                var current = groupStates[ix];
                if (current === undefined)
                    continue;
                currentState = copyHash(current, "AggregateState");
            }
        }
        var resolved = this.resolve(prefix);
        var start = prefixRound;
        groupStates[prefixRound] = currentState;
        if (!currentState) {
            currentState = groupStates[prefixRound] = op(this.newResultState(), resolved);
            var cur = this.getResultPrefix(prefix, this.stateToResult(currentState), 1);
            results.push(cur);
            start = prefixRound + 1;
        }
        for (var ix = start, len = Math.max(groupStates.length, prefixRound + 1); ix < len; ix++) {
            var current = groupStates[ix];
            if (current === undefined)
                continue;
            var prevResult = this.getResultPrefix(prefix, this.stateToResult(current), -1);
            current = groupStates[prefixRound] = op(current, resolved);
            var neueResult = this.getResultPrefix(prefix, this.stateToResult(current), 1);
            results.push(prevResult);
            results.push(neueResult);
        }
        return true;
    };
    return AggregateNode;
}(Node));
exports.AggregateNode = AggregateNode;
//------------------------------------------------------------------------------
// SortNode
//------------------------------------------------------------------------------
var SortNode = (function (_super) {
    __extends(SortNode, _super);
    // @TODO: allow for multiple returns
    function SortNode(groupRegisters, projectRegisters, directions, resultRegisters) {
        var _this = _super.call(this) || this;
        _this.groupRegisters = groupRegisters;
        _this.projectRegisters = projectRegisters;
        _this.directions = directions;
        _this.resultRegisters = resultRegisters;
        _this.name = "Sort";
        _this.traceType = trace_1.TraceNode.Aggregate;
        _this.groups = {};
        _this.resolved = [];
        _this.resolvedDirections = [];
        _this.add = function (state, resolved, round, results) {
            var resultPrefix = _this.resultPrefix;
            var neue = copyArray(resolved, "SortIntermediate");
            var ix = 0;
            for (var _i = 0, _a = state.sorted; _i < _a.length; _i++) {
                var item = _a[_i];
                if (_this.isGreater(item, resolved)) {
                    break;
                }
                ix++;
            }
            var outOffset = _this.resultRegisters[0].offset;
            state.sorted.splice(ix, 0, neue);
            results.push(resultPrefix(neue, outOffset, ix, round, 1));
            ix++;
            for (; ix < state.sorted.length; ix++) {
                var cur = state.sorted[ix];
                results.push(resultPrefix(cur, outOffset, ix - 1, round, -1));
                results.push(resultPrefix(cur, outOffset, ix, round, 1));
            }
            return state;
        };
        _this.remove = function (state, resolved, round, results) {
            var resultPrefix = _this.resultPrefix;
            var ix = 0;
            var found = false;
            for (var _i = 0, _a = state.sorted; _i < _a.length; _i++) {
                var item = _a[_i];
                if (_this.prefixEqual(item, resolved)) {
                    break;
                }
                ix++;
            }
            state.sorted.splice(ix, 1);
            var outOffset = _this.resultRegisters[0].offset;
            results.push(resultPrefix(resolved, outOffset, ix, round, -1));
            for (; ix < state.sorted.length; ix++) {
                var cur = state.sorted[ix];
                results.push(resultPrefix(cur, outOffset, ix + 1, round, -1));
                results.push(resultPrefix(cur, outOffset, ix, round, 1));
            }
            return state;
        };
        _this.groupKey = IntermediateIndex.CreateKeyFunction(groupRegisters);
        _this.projectKey = IntermediateIndex.CreateKeyFunction(projectRegisters);
        _this.sortRegisters = groupRegisters.concat(projectRegisters).filter(isRegister);
        return _this;
    }
    SortNode.prototype.groupPrefix = function (group, prefix) {
        var projection = this.projectKey(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var prefixCount = prefix[prefix.length - 1];
        var delta = 0;
        var found = this.groups[group];
        if (!found) {
            found = this.groups[group] = { result: [] };
        }
        var counts = found[projection] || [];
        var totalCount = 0;
        var countIx = 0;
        for (var _i = 0, counts_2 = counts; _i < counts_2.length; _i++) {
            var count = counts_2[_i];
            // we need the total up to our current round
            if (countIx > prefixRound)
                break;
            countIx++;
            if (!count)
                continue;
            totalCount += count;
        }
        if (totalCount && totalCount + prefixCount <= 0) {
            // subtract
            delta = -1;
        }
        else if (totalCount === 0 && totalCount + prefixCount > 0) {
            // add
            delta = 1;
        }
        else if (totalCount + prefixCount < 0) {
            // we have removed more values than exist?
            throw new Error("Negative total count for an aggregate projection");
        }
        else {
            // otherwise this change doesn't impact the projected count, we've just added
            // or removed a support.
        }
        counts[prefixRound] = (counts[prefixRound] || 0) + prefixCount;
        found[projection] = counts;
        return delta;
    };
    SortNode.prototype.resolve = function (prefix) {
        var _a = this, resolved = _a.resolved, resolvedDirections = _a.resolvedDirections;
        if (resolved.length < prefix.length)
            resolved.length = prefix.length;
        for (var _i = 0, _b = this.sortRegisters; _i < _b.length; _i++) {
            var field = _b[_i];
            if (isRegister(field)) {
                resolved[field.offset] = prefix[field.offset];
            }
        }
        var ix = 0;
        for (var _c = 0, _d = this.directions; _c < _d.length; _c++) {
            var field = _d[_c];
            if (isRegister(field)) {
                resolvedDirections[ix] = exports.GlobalInterner.reverse(prefix[field.offset]);
            }
            else {
                resolvedDirections[ix] = exports.GlobalInterner.reverse(field);
            }
            ix++;
        }
        return resolved;
    };
    SortNode.prototype.isGreater = function (a, b) {
        var resolvedDirections = this.resolvedDirections;
        var dirIx = 0;
        var dir = resolvedDirections[dirIx] || "up";
        for (var _i = 0, _a = this.projectRegisters; _i < _a.length; _i++) {
            var register = _a[_i];
            var offset = register.offset;
            var aV = exports.GlobalInterner.reverse(a[offset]);
            var bV = exports.GlobalInterner.reverse(b[offset]);
            if ((dir === "up" && aV > bV) ||
                (dir === "down" && aV < bV)) {
                return dir;
            }
            else if (aV !== bV) {
                return false;
            }
            dirIx++;
            dir = resolvedDirections[dirIx] || dir;
        }
        return false;
    };
    SortNode.prototype.prefixEqual = function (a, b) {
        var ix = -1;
        for (var _i = 0, a_2 = a; _i < a_2.length; _i++) {
            var field = a_2[_i];
            ix++;
            if (field !== b[ix]) {
                return false;
            }
        }
        return true;
    };
    SortNode.prototype.exec = function (context, input, prefix, transaction, round, results, changes) {
        var group = this.groupKey(prefix);
        var prefixRound = prefix[prefix.length - 2];
        var delta = this.groupPrefix(group, prefix);
        var op = this.add;
        if (!delta)
            return false;
        if (delta < 0)
            op = this.remove;
        var groupStates = this.groups[group].result;
        var currentState = groupStates[prefixRound];
        if (!currentState) {
            // otherwise we have to find the most recent result that we've seen
            for (var ix = 0, len = Math.min(groupStates.length, prefixRound); ix < len; ix++) {
                var current = groupStates[ix];
                if (current === undefined)
                    continue;
                currentState = copyHash(current, "SortState");
            }
        }
        var resolved = this.resolve(prefix);
        var start = prefixRound;
        groupStates[prefixRound] = currentState;
        if (!currentState) {
            currentState = groupStates[prefixRound] = op(this.newResultState(), resolved, prefixRound, results);
            start = prefixRound + 1;
        }
        for (var ix = start, len = Math.max(groupStates.length, prefixRound + 1); ix < len; ix++) {
            var current = groupStates[ix];
            if (current === undefined)
                continue;
            current = groupStates[prefixRound] = op(current, resolved, ix, results);
        }
        return true;
    };
    SortNode.prototype.resultPrefix = function (prefix, outOffset, pos, round, count) {
        var item = copyArray(prefix, "SortResult");
        // add one here because we're one indexed
        item[outOffset] = exports.GlobalInterner.intern(pos + 1);
        item[item.length - 2] = round;
        item[item.length - 1] = count;
        return item;
    };
    SortNode.prototype.newResultState = function () {
        return { sorted: [], sortLookup: {} };
    };
    return SortNode;
}(Node));
exports.SortNode = SortNode;
//------------------------------------------------------------------------------
// Block
//------------------------------------------------------------------------------
var Block = (function () {
    function Block(name, nodes, totalRegisters) {
        this.name = name;
        this.nodes = nodes;
        this.totalRegisters = totalRegisters;
        this.results = new Iterator();
        this.initial = createArray();
        for (var ix = 0; ix < this.totalRegisters + 2; ix++) {
            this.initial[ix] = undefined;
        }
    }
    Block.prototype.toString = function () {
        var content = this.nodes.map(toString).join(",\n");
        return "Block(\"" + this.name + "\", [\n  " + indent(content, 2) + "\n])";
    };
    Block.prototype.exec = function (context, input, transaction) {
        this.results.clear();
        this.results.push(this.initial.slice());
        var prefix;
        var iter = this.results;
        for (var _i = 0, _a = this.nodes; _i < _a.length; _i++) {
            var node = _a[_i];
            node.results.clear();
            if (iter.length === 0) {
                if (node instanceof AntiJoin) {
                    node.exec(context, input, this.initial.slice(), transaction.transaction, transaction.round, node.results, transaction);
                    iter = node.results.iter();
                }
            }
            else {
                while ((prefix = iter.next()) !== undefined) {
                    context.tracer.node(node, prefix);
                    node.exec(context, input, prefix, transaction.transaction, transaction.round, node.results, transaction);
                    context.tracer.pop(trace_1.TraceFrameType.Node);
                }
                iter = node.results.iter();
            }
        }
        return true;
    };
    return Block;
}());
exports.Block = Block;
//------------------------------------------------------------------------------
// EvaluationContext
//------------------------------------------------------------------------------
var EvaluationContext = (function () {
    function EvaluationContext(index) {
        this.index = index;
        this.distinctIndex = new indexes_1.DistinctIndex();
        this.intermediates = {};
        this.exportIndex = {};
        this.tracer = TRACE ? new trace_1.Tracer(this) : new trace_1.NoopTracer(this);
    }
    return EvaluationContext;
}());
exports.EvaluationContext = EvaluationContext;
var Transaction = (function () {
    function Transaction(context, transaction, blocks, exportHandler) {
        this.context = context;
        this.transaction = transaction;
        this.blocks = blocks;
        this.exportHandler = exportHandler;
        this.round = -1;
        this.changes = [];
        this.lastFrame = 0;
        this.outputs = new Iterator();
        this.roundChanges = [];
        this.frameCommits = [];
        this.framePartialCommits = [];
        this.exportedChanges = {};
        context.tracer.transaction(transaction);
    }
    Transaction.prototype.output = function (context, change) {
        // debug("        E~", change.toString(), context.tracker.activeBlock);
        var outputs = this.outputs;
        var distinctIndex = context.distinctIndex, tracer = context.tracer;
        tracer.maybeOutput(change);
        outputs.clear();
        distinctIndex.distinct(change, outputs);
        tracer.postDistinct();
        outputs.reset();
        var output;
        while (output = outputs.next()) {
            tracer.output(output);
            // debug("          <-", output.toString())
            var cur = this.roundChanges[output.round] || createArray("roundChangesArray");
            cur.push(output);
            this.roundChanges[output.round] = cur;
        }
        tracer.pop(trace_1.TraceFrameType.MaybeOutput);
    };
    Transaction.prototype.commit = function (context, change) {
        context.tracer.commit(change);
        var outputs = this.outputs;
        if (change instanceof RemoveVsChange) {
            this.framePartialCommits.push(change);
        }
        else {
            this.frameCommits.push(change);
        }
        // debug("          <-!", change.toString())
    };
    Transaction.prototype.export = function (context, blockId, change) {
        var e = change.e, a = change.a, v = change.v, count = change.count;
        var beav = blockId + "|" + e + "|" + a + "|" + v;
        var old = context.exportIndex[beav] || 0;
        var neue = old + count;
        var delta = 0;
        // Once you go negative you don't go back.
        if (old === 0 && neue > 0)
            delta = 1;
        else if (old > 0 && neue === 0)
            delta = -1;
        context.exportIndex[beav] = neue;
        if (delta) {
            var exportedChange = new Change(e, a, v, change.n, this.transaction, 0, delta);
            if (!this.exportedChanges[blockId])
                this.exportedChanges[blockId] = [exportedChange];
            else
                this.exportedChanges[blockId].push(exportedChange);
        }
    };
    Transaction.prototype.prepareRound = function (context, changeIx) {
        var _a = this, roundChanges = _a.roundChanges, changes = _a.changes;
        var next = changes[changeIx];
        var maxRound = roundChanges.length;
        var oldLength = changes.length;
        if (!next && this.round < maxRound) {
            for (var ix = this.round + 1; ix < maxRound; ix++) {
                var nextRoundChanges = roundChanges[ix];
                if (nextRoundChanges) {
                    this.collapseMultiplicity(nextRoundChanges, changes);
                    // If we've got new changes to go through, we're done
                    if (oldLength < changes.length)
                        return;
                }
            }
        }
        var _b = this, frameCommits = _b.frameCommits, framePartialCommits = _b.framePartialCommits;
        if (!next && (frameCommits.length || framePartialCommits.length)) {
            for (var _i = 0, framePartialCommits_1 = framePartialCommits; _i < framePartialCommits_1.length; _i++) {
                var commit = framePartialCommits_1[_i];
                commit.toRemoveChanges(context, frameCommits);
            }
            var collapsedCommits = [];
            this.collapseCommits(this.frameCommits, collapsedCommits);
            var collapsedChanges = [];
            this.collapseMultiplicity(collapsedCommits, collapsedChanges);
            if (collapsedChanges.length) {
                context.tracer.frame(collapsedChanges);
                this.lastFrame = this.changes.length;
                this.round = -1;
                this.roundChanges = [];
                this.frameCommits = [];
                this.framePartialCommits = [];
                for (var _c = 0, collapsedChanges_1 = collapsedChanges; _c < collapsedChanges_1.length; _c++) {
                    var commit = collapsedChanges_1[_c];
                    if (commit.count > 0)
                        commit.count = Infinity;
                    else if (commit.count < 0)
                        commit.count = -Infinity;
                    // debug("    ->! ", commit.toString())
                    this.output(context, commit);
                }
                this.prepareRound(context, changeIx);
                // debug(" ---------------- NEW FRAME -------------------")
            }
        }
    };
    Transaction.prototype.collapseCommits = function (changes, results /* output */) {
        // We sort the changes to group all the same EAVs together.
        changes.sort(function (a, b) {
            var nodeDiff = a.n - b.n;
            if (!nodeDiff) {
                var eDiff = a.e - b.e;
                if (!eDiff) {
                    var aDiff = a.a - b.a;
                    if (!aDiff) {
                        var vDiff = a.v - b.v;
                        return vDiff;
                    }
                    return aDiff;
                }
                return eDiff;
            }
            return nodeDiff;
        });
        var changeIx = 0;
        for (var changeIx_1 = 0; changeIx_1 < changes.length; changeIx_1++) {
            var current = changes[changeIx_1];
            var currentType = current instanceof RemoveChange ? true : false;
            if (currentType) {
                current = new RemoveChange(current.e, current.a, current.v, current.n, current.transaction, current.round, current.count);
            }
            else {
                current = new Change(current.e, current.a, current.v, current.n, current.transaction, current.round, current.count);
            }
            // Collapse each subsequent matching EAV's multiplicity into the current one's.
            while (changeIx_1 + 1 < changes.length) {
                var next = changes[changeIx_1 + 1];
                if (current.n === next.n && next.e == current.e && next.a == current.a && next.v == current.v) {
                    current.count += next.count;
                    changeIx_1++;
                }
                else {
                    break;
                }
            }
            current.round = 0;
            if (currentType && current.count < 0) {
                results.push(current);
            }
            else if (!currentType && current.count > 0) {
                results.push(current);
            }
        }
        return results;
    };
    Transaction.prototype.collapseMultiplicity = function (changes, results /* output */, createNew) {
        if (createNew === void 0) { createNew = false; }
        // We sort the changes to group all the same EAVs together.
        changes.sort(function (a, b) {
            var eDiff = a.e - b.e;
            if (!eDiff) {
                var aDiff = a.a - b.a;
                if (!aDiff) {
                    return a.v - b.v;
                }
                return aDiff;
            }
            return eDiff;
        });
        var changeIx = 0;
        for (var changeIx_2 = 0; changeIx_2 < changes.length; changeIx_2++) {
            var current = changes[changeIx_2];
            if (createNew) {
                current = new Change(current.e, current.a, current.v, current.n, current.transaction, current.round, current.count);
            }
            // Collapse each subsequent matching EAV's multiplicity into the current one's.
            while (changeIx_2 + 1 < changes.length) {
                var next = changes[changeIx_2 + 1];
                if (next.e == current.e && next.a == current.a && next.v == current.v) {
                    current.count += next.count;
                    changeIx_2++;
                }
                else {
                    break;
                }
            }
            if (current.count !== 0)
                results.push(current);
        }
        return results;
    };
    Transaction.prototype.exec = function (context) {
        var _a = this, changes = _a.changes, roundChanges = _a.roundChanges;
        var index = context.index, tracer = context.tracer;
        tracer.frame([]);
        var total = 0;
        var frames = 0;
        var changeIx = 0;
        this.prepareRound(context, changeIx);
        while (changeIx < changes.length) {
            var change = changes[changeIx];
            tracer.input(change);
            total++;
            if (total > 10000) {
                console.error("bad");
                break;
            }
            if (this.round !== 0 && change.round === 0) {
                frames++;
                if (frames > 10) {
                    console.error("Failed to terminate");
                    break;
                }
            }
            this.round = change.round;
            // debug("Round:", this.round);
            // debug("  <- ", change.toString())
            for (var _i = 0, _b = this.blocks; _i < _b.length; _i++) {
                var block = _b[_i];
                tracer.block(block.name);
                //debug("    ", block.name);
                block.exec(context, change, this);
                tracer.pop(trace_1.TraceFrameType.Block);
            }
            // debug("");
            index.insert(change);
            tracer.pop(trace_1.TraceFrameType.Input);
            changeIx++;
            this.prepareRound(context, changeIx);
        }
        var exportingBlocks = Object.keys(this.exportedChanges);
        if (exportingBlocks.length) {
            if (!this.exportHandler)
                throw new Error("Unable to export changes without export handler.");
            for (var _c = 0, exportingBlocks_1 = exportingBlocks; _c < exportingBlocks_1.length; _c++) {
                var blockId = exportingBlocks_1[_c];
                var exports_1 = createArray("exportsArray");
                this.collapseMultiplicity(this.exportedChanges[+blockId], exports_1);
                this.exportedChanges[+blockId] = exports_1;
            }
            try {
                this.exportHandler(this.exportedChanges);
            }
            catch (e) {
                tracer.pop(trace_1.TraceFrameType.Transaction);
                throw e;
            }
        }
        // Once the transaction is effectively done, we need to clean up after ourselves. We
        // arena allocated a bunch of IDs related to function call outputs, which we can now
        // safely release.
        exports.GlobalInterner.releaseArena("functionOutput");
        tracer.pop(trace_1.TraceFrameType.Transaction);
    };
    return Transaction;
}());
exports.Transaction = Transaction;
var BlockChangeTransaction = (function (_super) {
    __extends(BlockChangeTransaction, _super);
    function BlockChangeTransaction(context, transaction, added, removed, blocks, exportHandler) {
        var _this = _super.call(this, context, transaction, blocks, exportHandler) || this;
        _this.context = context;
        _this.transaction = transaction;
        _this.added = added;
        _this.removed = removed;
        _this.blocks = blocks;
        _this.exportHandler = exportHandler;
        return _this;
    }
    BlockChangeTransaction.prototype.exec = function (context) {
        // To remove a block, we run a change with negative count through the system that
        // is meant to compute all the results the block has generated and remove them.
        for (var _i = 0, _a = this.removed; _i < _a.length; _i++) {
            var remove = _a[_i];
            remove.exec(context, BLOCK_REMOVE, this);
        }
        // To add a block, we do the same as remove, but with a positive change that causes
        // the block to compute all results based on the state of the indexes
        for (var _b = 0, _c = this.added; _b < _c.length; _b++) {
            var add = _c[_b];
            add.exec(context, BLOCK_ADD, this);
        }
        _super.prototype.exec.call(this, context);
    };
    return BlockChangeTransaction;
}(Transaction));
exports.BlockChangeTransaction = BlockChangeTransaction;
// window["counts"] = {};
//# sourceMappingURL=runtime.js.map